# copyright (c) 2020 PaddlePaddle Authors. All Rights Reserve.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
This code is refer from:
https://github.com/WenmuZhou/DBNet.pytorch/blob/master/data_loader/modules/make_border_map.py
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

import numpy as np
import cv2

np.seterr(divide='ignore', invalid='ignore')
import pyclipper
from shapely.geometry import Polygon
import sys
import warnings

warnings.simplefilter("ignore")

__all__ = ['MakeBorderMap']


class MakeBorderMap(object):
    """ 计算文本区域阈值图标签类

    ppocr db模型代码阅读理解_垚1234的博客-CSDN博客:
    https://blog.csdn.net/weixin_41393000/article/details/118740663
    """
    def __init__(self,
                 shrink_ratio=0.4,
                 thresh_min=0.3,
                 thresh_max=0.7,
                 **kwargs):
        self.shrink_ratio = shrink_ratio
        self.thresh_min = thresh_min
        self.thresh_max = thresh_max

    def __call__(self, data):
        img = data['image']
        text_polys = data['polys']
        ignore_tags = data['ignore_tags']

        # 1. 生成空模版
        canvas = np.zeros(img.shape[:2], dtype=np.float32)
        mask = np.zeros(img.shape[:2], dtype=np.float32)

        for i in range(len(text_polys)):
            if ignore_tags[i]:
                continue

            # 2. draw_border_map函数根据解码后的box信息计算阈值图标签
            self.draw_border_map(text_polys[i], canvas, mask=mask)
        canvas = canvas * (self.thresh_max - self.thresh_min) + self.thresh_min

        data['threshold_map'] = canvas
        data['threshold_mask'] = mask

        # from pyxllib.ai.paddlelib import show_feature_map
        # show_feature_map([[data['image'][:,:,0].astype('float32') / 255, canvas, mask]])
        return data

    def draw_border_map(self, polygon, canvas, mask):
        polygon = np.array(polygon)
        assert polygon.ndim == 2
        assert polygon.shape[1] == 2

        polygon_shape = Polygon(polygon)
        if polygon_shape.area <= 0:
            return
        # 多边形内缩
        distance = polygon_shape.area * (
            1 - np.power(self.shrink_ratio, 2)) / polygon_shape.length
        subject = [tuple(l) for l in polygon]
        padding = pyclipper.PyclipperOffset()
        padding.AddPath(subject, pyclipper.JT_ROUND, pyclipper.ET_CLOSEDPOLYGON)
        # 计算mask
        padded_polygon = np.array(padding.Execute(distance)[0])
        cv2.fillPoly(mask, [padded_polygon.astype(np.int32)], 1.0)

        xmin = padded_polygon[:, 0].min()
        xmax = padded_polygon[:, 0].max()
        ymin = padded_polygon[:, 1].min()
        ymax = padded_polygon[:, 1].max()
        width = xmax - xmin + 1
        height = ymax - ymin + 1

        polygon[:, 0] = polygon[:, 0] - xmin
        polygon[:, 1] = polygon[:, 1] - ymin

        xs = np.broadcast_to(
            np.linspace(
                0, width - 1, num=width).reshape(1, width), (height, width))
        ys = np.broadcast_to(
            np.linspace(
                0, height - 1, num=height).reshape(height, 1), (height, width))

        distance_map = np.zeros(
            (polygon.shape[0], height, width), dtype=np.float32)
        for i in range(polygon.shape[0]):
            j = (i + 1) % polygon.shape[0]
            # 计算点到线的距离
            absolute_distance = self._distance(xs, ys, polygon[i], polygon[j])
            distance_map[i] = np.clip(absolute_distance / distance, 0, 1)
        distance_map = distance_map.min(axis=0)

        xmin_valid = min(max(0, xmin), canvas.shape[1] - 1)
        xmax_valid = min(max(0, xmax), canvas.shape[1] - 1)
        ymin_valid = min(max(0, ymin), canvas.shape[0] - 1)
        ymax_valid = min(max(0, ymax), canvas.shape[0] - 1)
        canvas[ymin_valid:ymax_valid + 1, xmin_valid:xmax_valid + 1] = np.fmax(
            1 - distance_map[ymin_valid - ymin:ymax_valid - ymax + height,
                             xmin_valid - xmin:xmax_valid - xmax + width],
            canvas[ymin_valid:ymax_valid + 1, xmin_valid:xmax_valid + 1])

    def _distance(self, xs, ys, point_1, point_2):
        '''
        compute the distance from point to a line
        ys: coordinates in the first axis
        xs: coordinates in the second axis
        point_1, point_2: (x, y), the end of the line
        '''
        height, width = xs.shape[:2]
        square_distance_1 = np.square(xs - point_1[0]) + np.square(ys - point_1[
            1])
        square_distance_2 = np.square(xs - point_2[0]) + np.square(ys - point_2[
            1])
        square_distance = np.square(point_1[0] - point_2[0]) + np.square(
            point_1[1] - point_2[1])

        cosin = (square_distance - square_distance_1 - square_distance_2) / (
            2 * np.sqrt(square_distance_1 * square_distance_2))
        square_sin = 1 - np.square(cosin)
        square_sin = np.nan_to_num(square_sin)
        result = np.sqrt(square_distance_1 * square_distance_2 * square_sin /
                         square_distance)

        result[cosin <
               0] = np.sqrt(np.fmin(square_distance_1, square_distance_2))[cosin
                                                                           < 0]
        # self.extend_line(point_1, point_2, result)
        return result

    def extend_line(self, point_1, point_2, result, shrink_ratio):
        ex_point_1 = (int(
            round(point_1[0] + (point_1[0] - point_2[0]) * (1 + shrink_ratio))),
                      int(
                          round(point_1[1] + (point_1[1] - point_2[1]) * (
                              1 + shrink_ratio))))
        cv2.line(
            result,
            tuple(ex_point_1),
            tuple(point_1),
            4096.0,
            1,
            lineType=cv2.LINE_AA,
            shift=0)
        ex_point_2 = (int(
            round(point_2[0] + (point_2[0] - point_1[0]) * (1 + shrink_ratio))),
                      int(
                          round(point_2[1] + (point_2[1] - point_1[1]) * (
                              1 + shrink_ratio))))
        cv2.line(
            result,
            tuple(ex_point_2),
            tuple(point_2),
            4096.0,
            1,
            lineType=cv2.LINE_AA,
            shift=0)
        return ex_point_1, ex_point_2



# Copyright (c) 2020 PaddlePaddle Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import paddle
import paddle.nn as nn

from .rec_ctc_loss import CTCLoss
from .center_loss import CenterLoss
from .ace_loss import ACELoss

from .distillation_loss import DistillationCTCLoss
from .distillation_loss import DistillationDMLLoss
from .distillation_loss import DistillationDistanceLoss, DistillationDBLoss, DistillationDilaDBLoss


class CombinedLoss(nn.Layer):
    """
    CombinedLoss:
        a combionation of loss function
    """

    def __init__(self, loss_config_list=None):
        super().__init__()
        self.loss_func = []
        self.loss_weight = []
        assert isinstance(loss_config_list, list), (
            'operator config should be a list')
        for config in loss_config_list:
            assert isinstance(config,
                              dict) and len(config) == 1, "yaml format error"
            name = list(config)[0]
            param = config[name]
            assert "weight" in param, "weight must be in param, but param just contains {}".format(
                param.keys())
            self.loss_weight.append(param.pop("weight"))
            self.loss_func.append(eval(name)(**param))

    def forward(self, input, batch, **kargs):
        loss_dict = {}
        loss_all = 0.
        for idx, loss_func in enumerate(self.loss_func):
            loss = loss_func(input, batch, **kargs)
            if isinstance(loss, paddle.Tensor):
                loss = {"loss_{}_{}".format(str(loss), idx): loss}

            weight = self.loss_weight[idx]

            loss = {key: loss[key] * weight for key in loss}

            if "loss" in loss:
                loss_all += loss["loss"]
            else:
                loss_all += paddle.add_n(list(loss.values()))
            loss_dict.update(loss)
        loss_dict["loss"] = loss_all
        return loss_dict



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# @Author : 陈坤泽
# @Email  : 877362867@qq.com
# @Date   : 2022/05/05 16:44


from collections import defaultdict
import math
import re

from pyxllib.prog.pupil import run_once

# 中文wiki颜色列表：https://zh.m.wikipedia.org/zh-hans/%E9%A2%9C%E8%89%B2%E5%88%97%E8%A1%A8
_COLOR_LIST0 = """
000000,黑色,Black
000080,藏青,Navy Blue
00008B,暗蓝,Dark Blue
0000CD,中蓝,Medium Blue
0000FF,蓝色,Blue
002FA7,国际奇连蓝,International Klein Blue
003153,普鲁士蓝,Prussian blue
003399,暗粉蓝,Dark Powder Blue
0033FF,极浓海蓝,Ultramarine
00477D,水手蓝,Marine Blue
0047AB,钴蓝,Cobalt Blue
004D99,矿蓝,Mineral Blue
006374,浓蓝,Strong Blue
006400,暗绿,Dark Green
007FFF,湛蓝,Azure
008000,绿色,Green
008080,凫绿,Teal
00808C,孔雀蓝,Peacock Blue
008B8B,暗青,Dark Cyan
00A15C,孔雀绿,Peacock Green
00BFFF,深天蓝,Deep Sky Blue
00CED1,暗绿松石色,Dark Turquoise
00FA9A,中春绿色,Medium Spring Green
00FF00,鲜绿色,Lime
00FF80,春绿,Spring Green
00FFEF,土耳其蓝,Turquoise Blue
00FFFF,青色,Cyan
082567,蓝宝石色,Sapphire
0DBF8C,青蓝,Cyan Blue
127436,铬绿,Viridian
16982B,薄荷绿,Mint
191970,午夜蓝,Midnight Blue
1E90FF,道奇蓝,Dodger Blue
20B2AA,亮海绿,Light Sea Green
228B22,森林绿,Forest Green
22C32E,孔雀石绿,Malachite
24367D,暗矿蓝,Dark Mineral Blue
26619C,天青石蓝,Lapis Lazuli
2A52BE,蔚蓝,Cerulean blue
2E8B57,海绿,Sea Green
2F4F4F,暗岩灰,Dark Slate Gray
32CD32,柠檬绿,Lime Green
36BF36,常春藤绿,Ivy Green
3CB371,中海绿,Medium Sea Green
40E0D0,绿松石色,Turquoise
4169E1,品蓝,Royal Blue
4682B4,钢青色,Steel Blue
4798B3,萨克斯蓝,Saxe Blue
483D8B,暗岩蓝,Dark Slate Blue
48D1CC,中绿松石色,Medium Turquoise
4B0080,靛色,Indigo
4D1F00,椰褐,Coconut Brown
4D3900,咖啡色,Coffee
4D80E6,鼠尾草蓝,Salvia Blue
4DE680,绿松石绿,Turquoise Green
50C878,碧绿,Emerald
556B2F,暗橄榄绿,Dark Olive Green
5686BF,韦奇伍德瓷蓝,Wedgwood Blue
5E86C1,灰丁宁蓝,Pale Denim
5F9EA0,军服蓝,Cadet Blue
625B57,铁灰色,Iron Gray
6495ED,矢车菊蓝,Cornflower Blue
66CDAA,中碧蓝色,Medium Aquamarine
66FF00,明绿,Bright Green
66FF59,钴绿,Cobalt Green
66FFE6,水蓝,Aqua Blue
696969,昏灰,Dim Gray
697723,苔藓绿,Moss Green
6A0DAD,紫色,Purple
6A5ACD,岩蓝,Slate Blue
6B8E23,橄榄军服绿,Olive Drab
704214,乌贼墨色,Sepia
708090,岩灰,Slate Gray
73B839,叶绿,Foliage Green
73E68C,孔雀石绿,Very Light Malachite Green
7400A1,三色堇紫,Pansy
778899,亮岩灰,Light Slate Gray
7B68EE,中岩蓝,Medium Slate Blue
7CFC00,草坪绿,Lawn Green
7F00FF,紫罗兰色,Violet
7FFF00,查特酒绿,Chartreuse
7FFFD4,碧蓝色,Aquamarine
800000,栗色,Maroon
800020,勃艮第酒红,Burgundy
800080,宗主教,Patriarch
808000,橄榄色,Olive
808080,灰色,Gray
8674A1,浅灰紫红,Grayish Purple
87CEEB,天空蓝,Sky Blue
87CEFA,浅天蓝,Light Sky Blue
89CFF0,浅蓝,Baby Blue
8A2BE2,蓝紫,Blue Violet
8B0000,暗红,Dark Red
8B008B,暗洋红,Dark Magenta
8B4513,鞍褐,Saddle Brown
8CE600,苹果绿,Apple Green
8E4585,梅红色,Plum
8FBC8F,暗海绿,Dark Sea Green
90EE90,亮绿,Light Green
9370DB,中紫红,Medium Purple
9400D3,暗紫,Dark Violet
98FB98,灰绿,Pale Green
990036,枢机红,Cardinal Red
9932CC,暗兰紫,Dark Orchid
9966CC,紫水晶色,Amethyst
996B1F,卡其色,Khaki
99E64D,草绿,Grass Green
99FF4D,嫩绿,Fresh Leaves
9ACD32,黄绿,Yellow Green
A0522D,赭黄,Sienna
A16B47,驼色,Camel
A39DAE,矿紫,Mineral Violet
A52A2A,褐色,Brown
A9A9A9,暗灰,Dark Gray
ACE1AF,青瓷绿,Celadon
ADD8E6,亮蓝,Light Blue
ADFF2F,绿黄,Green Yellow
AFDFE4,水色,Aqua
AFEEEE,灰绿松石色,Pale Turquoise
B09DB9,亮紫,Light Violet
B0C4DE,亮钢蓝,Light Steel Blue
B0E0E6,粉蓝,Powder Blue
B22222,砖红色,Fire Brick
B57EDC,薰衣草紫,Lavender
B784A7,优品紫红,Opera Mauve
B8860B,暗金菊色,Dark Goldenrod
B8DDC8,苍色,Horizon Blue
BA55D3,中兰紫,Medium Orchid
BC8F8F,玫瑰褐,Rosy Brown
BDB76B,暗卡其色,Dark Khaki
C08081,陈玫红,Old Rose
C0C0C0,银色,Silver
C3CDE6,长春花色,Periwinkle
C71585,中青紫红,Medium Violet Red
C8A2C8,紫丁香色,Lilac
C9A0DC,紫藤色,Wisteria
CC0080,红宝石色,Ruby
CC5500,燃橙,Burnt Orange
CC7722,赭色,Ocher
CCA3CC,铁线莲紫,Clematis
CCB38C,灰土色,Pale Ocre
CCCC4D,芥末黄,Mustard
CCCCFF,长春花色,Lavender Blue
CCFF00,亮柠檬绿,Light Lime
CD5C5C,印度红,Indian Red
CD7F32,铜色,Bronze
CD853F,秘鲁色,Peru
D1EDF2,灰蓝,Pale Blue
D2691E,巧克力色,Chocolate
D2B48C,日晒色,Tan
D3D3D3,亮灰色,Light Gray
D8BFD8,蓟紫,Thistle
D94DFF,锦葵紫,Mallow
DA70D6,兰紫,Orchid
DAA520,金菊色,Goldenrod
DB7093,灰紫红,Pale Violet Red
DC143C,绯红,Crimson
DCDCDC,庚斯博罗灰,Gainsboro
DDA0DD,梅红色,Medium Lavender Magenta
DE3163,樱桃红,Cerise
DEB887,硬木色,Burly Wood
DF73FF,缬草紫,Heliotrope
E0B0FF,木槿紫,Mauve
E0FFFF,亮青,Light Cyan
E32636,茜红,Alizarin Crimson
E60000,鲜红,Strong Red
E6005C,胭脂红,Carmine
E63995,山茶红,Camellia
E68AB8,火鹤红,Flamingo
E69966,杏黄,Apricot
E6B800,铬黄,Chrome Yellow
E6C35C,茉莉黄,Jasmine
E6C3C3,沙棕,Sand Beige
E6CFE6,淡紫丁香色,Pail Lilac
E6D933,含羞草黄,Mimosa
E6E6FA,薰衣草紫,Lavender Mist
E9967A,暗鲑红,Dark Salmon
EE82EE,亮紫,Lavender Magenta
EEE8AA,灰金菊色,Pale Goldenrod
F08080,亮珊瑚色,Light Coral
F0E68C,亮卡其色,Light Khaki
F0F8FF,爱丽丝蓝,Alice Blue
F0FFF0,蜜瓜绿,Honeydew
F28500,橘色,Tangerine
F400A1,品红,Fuchsia
F4A460,沙褐,Sand Brown
F5DEB3,小麦色,Wheat
F5F5DC,米色,Beige
F5F5F5,白烟色,White Smoke
F5FFFA,薄荷奶油色,Mint Cream
F8F8FF,幽灵白,Ghost White
FA8072,鲑红,Salmon
FAEBD7,古董白,Antique White
FAF0E6,亚麻色,Linen
FAFAD2,亮金菊黄,Light Goldenrod Yellow
FDF5E6,旧蕾丝色,Old Lace
FF0000,红色,Red
FF007F,玫瑰红,Rose
FF00FF,洋红,Magenta
FF0DA6,洋玫瑰红,Magenta Rose
FF1493,深粉红,Deep Pink
FF2400,腥红,Scarlet
FF4500,橙红,Orange Red
FF4D00,朱红,Vermilion
FF4D40,柿子橙,Persimmon
FF6347,蕃茄红,Tomato
FF66CC,浅玫瑰红,Rose Pink
FF69B4,暖粉红,Hot Pink
FF7300,阳橙,Sun Orange
FF73B3,尖晶石红,Spinel Red
FF7F50,珊瑚红,Coral
FF8033,热带橙,Tropical Orange
FF8099,浅鲑红,Salmon Pink
FF80BF,浅珊瑚红,Coral Pink
FF8C00,暗橙,Dark Orange
FF9900,万寿菊黄,Marigold
FFA07A,亮鲑红,Light Salmon
FFA500,橙色,Orange
FFB366,蜜橙,Honey Orange
FFB3BF,壳黄红,Shell Pink
FFB3E6,浅珍珠红,Pearl Pink
FFB6C1,亮粉红,Light Pink
FFBF00,琥珀色,Amber
FFC0CB,粉红,Pink
FFD700,金色,Golden
FFD9E6,浅粉红,Baby Pink
FFDAB9,粉扑桃色,Peach Puff
FFDEAD,那瓦霍白,Navajo White
FFE4B5,鹿皮鞋色,Moccasin
FFE4C4,陶坯黄,Bisque
FFE4E1,雾玫瑰色,Misty Rose
FFE5B4,桃色,Peach
FFEBCD,杏仁白,Blanched Almond
FFEF00,鲜黄,Canary Yellow
FFEFD5,蕃木瓜色,Papaya Whip
FFF0F5,薰衣草紫红,Lavender Blush
FFF5EE,海贝色,Seashell
FFF8DC,玉米丝色,Cornsilk
FFFACD,柠檬绸色,Lemon Chiffon
FFFAF0,花卉白,Floral White
FFFAFA,雪色,Snow
FFFDD0,奶油色,Cream
FFFF00,黄色,Yellow
FFFF4D,月黄,Moon Yellow
FFFF99,香槟黄,Champagne Yellow
FFFFE0,亮黄,Light Yellow
FFFFF0,象牙色,Ivory
FFFFFF,白色,White
"""

# 自己标注英文颜色后，转出带中文名称的颜色
# 注意必须用\t做分隔符；而pycharm里默认会把\t替换成空格，所以可以配合Ctrl+D等技巧来操作
_COLOR_LIST1 = """
010203,浓黑,Rich black (FOGRA39)
36454F,木炭,Charcoal
F4F0EC,灰黄,Isabelline
E6FFED,git绿1,
CDFFD8,git绿2,
ACF2BD,git绿3,
FFEEF0,git红1,
FFDCE0,git红2,
FDB8C0,git红3,
F1F8FF,git蓝1,
DAECFE,git蓝2,
"""
# git绿1，github desktop，新增行三个级别绿，最浅的一种

_COLOR_LIST2 = """
000000,,Black
000080,,Navy blue
00009C,,Duke blue
0000CD,,Medium blue
0000FF,,Blue
000F89,,Phthalo blue
0014A8,,Zaffre
0018A8,,Blue (Pantone)
002147,,Oxford blue
002366,,Royal blue (dark)
002387,,Resolution blue
003153,,Prussian blue
0033AA,,UA blue
00356B,,Yale Blue
004040,,Rich black
00416A,,Indigo dye
004225,,British racing green
004242,,Warm black
0047AB,,Cobalt blue
0048BA,,Absolute Zero
004953,,Midnight green (eagle green)
004B49,,Deep jungle green
004F98,,USAFA blue
00563F,,Castleton green
00573F,,Gotham green
006400,,Dark green (X11)
006600,,Pakistan green
0067A5,,Sapphire blue
006A4E,,Bottle green
006B3C,,Cadmium green
006DB0,,Honolulu blue
00703C,,Dartmouth green
0070B8,,Spanish blue
0072BB,,French blue
007474,,Skobeloff
00755E,,Tropical rainforest
007AA5,,CG blue
007BA7,,Celadon blue
007BA7,,Cerulean
007BB8,,Star command blue
007F5C,,Spanish viridian
007F66,,Generic viridian
007FFF,,Azure
008000,,Ao (English)
008000,,Green (web)
008080,,Teal
0087BD,,Blue (NCS)
008B8B,,Dark cyan
009150,,Spanish green
0093AF,,Blue (Munsell)
009698,,Viridian green
009966,,Green-cyan
009B7D,,Paolo Veronese green
009E60,,Shamrock green
009F6B,,Green (NCS)
00A550,,Green (pigment)
00A693,,Persian green
00A86B,,Jade
00A877,,Green (Munsell)
00AB66,,GO green
00AD43,,Green (Pantone)
00B7EB,,Cyan (process)
00BFFF,,Capri
00BFFF,,Deep sky blue
00CC99,,Caribbean green
00CCCC,,Robin egg blue
00CCFF,,Vivid sky blue
00CED1,,Dark turquoise
00FA9A,,Medium spring green
00FF00,,Electric green
00FF00,,Green
00FF00,,Lime (web) (X11 green)
00FF40,,Erin
00FF7F,,Spring green
00FFCD,,Sea green (Crayola)
00FFEF,,Turquoise blue
00FFFF,,Aqua
00FFFF,,Cyan
00FFFF,,Spanish sky blue
010B13,,Rich black (FOGRA29)
013220,,Dark green
014421,,Forest green (traditional)
014421,,UP Forest green
01796F,,Pine green
0247FE,,Blue (RYB)
03C03C,,Dark pastel green
043927,,Sacramento State green
064E40,,Blue-green (color wheel)
0A7E8C,,Metallic Seaweed
0ABAB5,,Tiffany Blue
0BDA51,,Malachite
0D98BA,,Blue-green
0F52BA,,Sapphire
100C08,,Smoky black
1034A6,,Egyptian blue
1164B4,,Green-blue
123524,,Phthalo green
126180,,Blue sapphire
130a8f,,International Klein Blue
138808,,India green
1560BD,,Denim
15F4EE,,Fluorescent blue
16161D,,Eigengrau
177245,,Dark spring green
18453B,,MSU green
189BCC,,Weezy Blue
191970,,Midnight blue
195905,,Lincoln green
1974D2,,Bright navy blue
1974D2,,Navy blue (Crayola)
1A2421,,Dark jungle green
1B1811,,Black chocolate
1B1B1B,,Eerie black
1B4D3E,,Brunswick green
1B4D3E,,English green
1C05B3,,Trypan Blue
1C39BB,,Persian blue
1CA9C9,,Pacific blue
1CAC78,,Green (Crayola)
1D2951,,Space cadet
1DACD6,,Cerulean (Crayola)
1E90FF,,Dodger blue
1F75FE,,Blue (Crayola)
20B2AA,,Light sea green
2243B6,,Denim blue
228B22,,Forest green (web)
23297A,,St. Patrick's blue
232B2B,,Charleston green
242124,,Raisin black
246BCE,,Celtic blue
26428B,,Dark cornflower blue
26619C,,Lapis lazuli
27346F,,Astronaut
2887C8,,Green-blue (Crayola)
299617,,Slimy green
29AB87,,Jungle green
2A2F23,,Pine tree
2A52BE,,Cerulean blue
2D383A,,Outer space (Crayola)
2D5DA1,,Sapphire (Crayola)
2D68C4,,True Blue
2E2787,,Picotee blue
2E2D88,,Cosmic cobalt
2E5090,,YInMn Blue
2E5894,,B'dazzled blue
2E8B57,,Sea green
2F847C,,Celadon green
2a3439,,Gunmetal
301934,,Dark purple
30B21A,,Yellow-green (Color Wheel)
30BA8F,,Mountain Meadow
30BFBF,,Maximum blue green
317873,,Myrtle green
318CE7,,Bleu de France
319177,,Illuminating emerald
32127A,,Persian indigo
32174D,,Russian violet
324AB2,,Violet-blue
32CD32,,Lime green
333399,,Blue (pigment)
343434,,Jet
353839,,Onyx
354230,,Kombu green
355E3B,,Hunter green
36747D,,Ming
367588,,Teal blue
39A78E,,Zomp
39FF14,,Neon green
3AB09E,,Keppel
3B2F2F,,Black coffee
3B3C36,,Black olive
3B7A57,,Amazon
3C1414,,Dark sienna
3C341F,,Olive Drab 7
3C69E7,,Bluetiful
3CB371,,Medium sea green
3D0C02,,Black bean
3D2B1F,,Bistre
3DDC84,,Android green
3E8EDE,,Tufts blue
3EB489,,Mint
3F00FF,,Ultramarine
3FFF00,,Harlequin
40826D,,Viridian
40E0D0,,Turquoise
4166F5,,Ultramarine blue
4169E1,,Royal blue (light)
43302E,,Old burgundy
436B95,,Queen blue
43B3AE,,Verdigris
444C38,,Rifle green
446CCF,,Han blue
4666FF,,Neon blue
4682B4,,Steel blue
47ABCC,,Maximum blue
483C32,,Dark lava
483C32,,Taupe
483D8B,,Dark slate blue
48BF91,,Ocean green
48D1CC,,Medium turquoise
49796B,,Hooker's green
4A0000,,Oxblood
4A5D23,,Dark moss green
4A646C,,Deep Space Sparkle
4B0082,,Indigo
4B3621,,Café noir
4B5320,,Army green
4C2882,,Spanish violet
4C516D,,Independence
4C9141,,May green
4CBB17,,Kelly green
4D1A7F,,Blue-violet (color wheel)
4D5D53,,Feldgrau
4D8C57,,Middle green
4E5180,,Purple navy
4F42B5,,Ocean Blue
4F7942,,Fern green
5072A7,,Blue yonder
507D2A,,Sap green
50C878,,Emerald
50C878,,Paris Green
512888,,KSU purple
5218FA,,Han purple
534B4F,,Dark liver
536872,,Cadet
536878,,Dark electric blue
536878,,Payne's grey
543D37,,Dark liver (horses)
545AA7,,Liberty
54626F,,Black coral
555555,,Davy's grey
555D50,,Ebony
556B2F,,Dark olive green
563C5C,,English violet
568203,,Avocado
56887D,,Wintergreen Dream
56A0D3,,Carolina blue
58111A,,Chocolate Cosmos
58427C,,Cyber grape
59260B,,Seal brown
592720,,Caput mortuum
5946B2,,Plump Purple
5A4FCF,,Iris
5B3256,,Japanese violet
5B92E5,,United Nations blue
5D3954,,Dark byzantium
5DA493,,Polished Pine
5DADEC,,Blue jeans
5E8C31,,Maximum green
5F8A8B,,Steel Teal
5F9EA0,,Cadet blue
5FA777,,Forest green (Crayola)
5FA778,,Shiny Shamrock
6050DC,,Majorelle blue
6082B6,,Glaucous
612086,,Seance
614051,,Eggplant
635147,,Umber
644117,,Pullman Brown (UPS Brown)
6495ED,,Cornflower blue
65000B,,Rosewood
654321,,Dark brown
660000,,Blood red
66023C,,Tyrian purple
663399,,Rebecca Purple
664228,,Van Dyke brown
665D1E,,Antique bronze
666699,,Dark blue-gray
6699CC,,Blue-gray
6699CC,,Livid
66B032,,Green (RYB)
66DDAA,,Medium aquamarine
66FF00,,Bright green
66FF66,,Screamin' Green
673147,,Old mauve
673147,,Wine dregs
674846,,Rose ebony
674C47,,Liver
676767,,Granite gray
679267,,Russian green
682860,,Palatinate purple
683068,,Finn
696969,,Dim gray
6A0DAD,,Purple
6A5ACD,,Slate blue
6B4423,,Kobicha
6B8E23,,Olive Drab (3)
6C2E1F,,Liver (organ)
6C3082,,Eminence
6C541E,,Field drab
6CA0DC,,Little boy blue
6D9BC3,,Cerulean frost
6EAEA1,,Green Sheen
6F00FF,,Electric indigo
6F4E37,,Coffee
6F4E37,,Tuscan brown
701C1C,,Persian plum
701C1C,,Prune
702670,,Midnight
702963,,Byzantium
703642,,Catawba
704214,,Sepia
708090,,Slate gray
71A6D2,,Iceberg
722F37,,Wine
727472,,Nickel
72A0C1,,Air superiority blue
733380,,Maximum purple
7366BD,,Blue-violet (Crayola)
738678,,Xanadu
73C2FB,,Maya blue
74C365,,Mantis
757575,,Sonic silver
766EC8,,Violet-blue (Crayola)
76D7EA,,Sky blue (Crayola)
777696,,Rhythm
778899,,Light slate gray
778BA5,,Shadow blue
77B5FE,,French sky blue
78184A,,Pansy purple
7851A9,,Royal purple
79443B,,Bole
796878,,Old lavender
7B1113,,UP maroon
7B3F00,,Chocolate (traditional)
7B68EE,,Medium slate blue
7BB661,,Bud green
7C0A02,,Barn red
7C4848,,Tuscan red
7CB9E8,,Aero
7CFC00,,Lawn green
7DF9FF,,Electric blue
7E5E60,,Deep taupe
7ED4E6,,Middle blue
7F00FF,,Violet (color wheel)
7F1734,,Claret
7FFF00,,Chartreuse (web)
7FFFD4,,Aquamarine
800000,,Maroon (web)
800020,,Burgundy
800080,,Patriarch
800080,,Purple (web)
801818,,Falu red
80461B,,Russet
807532,,Spanish bistre
808000,,Olive
808080,,Gray (web)
81613C,,Coyote brown
826644,,Raw umber
838996,,Roman silver
841617,,OU Crimson red
841B2D,,Antique ruby
848482,,Battleship grey
848482,,Old silver
856088,,Chinese violet
856D4D,,French bistre
85754E,,Gold Fusion
8601AF,,Violet (RYB)
86608E,,French lilac
86608E,,Pomp and Power
86A1A9,,Tourmaline
87413F,,Brandy
87421F,,Fuzzy Wuzzy
87A96B,,Asparagus
87CEEB,,Sky blue
87CEFA,,Light sky blue
87FF2A,,Spring Frost
880085,,Mardi Gras
8806CE,,French violet
882D17,,Kobe
882D17,,Sienna
88540B,,Brown
893843,,Solid pink
893F45,,Cordovan
89CFF0,,Baby blue
8A2BE2,,Blue-violet
8A3324,,Burnt umber
8A496B,,Twilight lavender
8A795D,,Shadow
8A7F80,,Rocket metallic
8A9A5B,,Moss green
8A9A5B,,Turtle green
8B0000,,Dark red
8B008B,,Dark magenta
8B4513,,Saddle brown
8B72BE,,Middle blue purple
8B8589,,Taupe gray
8B8680,,Middle grey
8BA8B7,,Pewter Blue
8C92AC,,Cool grey
8CBED6,,Dark sky blue
8D4E85,,Razzmic Berry
8DA399,,Morning blue
8DB600,,Apple green
8DD9CC,,Middle blue green
8E3A59,,Quinacridone magenta
8E4585,,Plum
8F00FF,,Electric violet
8F00FF,,Violet
8F9779,,Artichoke
8FBC8F,,Dark sea green
8FD400,,Sheen green
905D5D,,Rose taupe
90EE90,,Light green
914E75,,Sugar Plum
915C83,,Antique fuchsia
915F6D,,Mauve taupe
915F6D,,Raspberry glace
91A3B0,,Cadet grey
922B3E,,Red-violet (Color wheel)
9370DB,,Medium purple
93C572,,Pistachio
93CCEA,,Light cornflower blue
9400D3,,Dark violet
954535,,Chestnut
960018,,Carmine
963D7F,,Violet (crayola)
967117,,Bistre brown
967117,,Drab
967117,,Mode beige
967117,,Sand dune
9678B6,,Purple mountain majesty
96C8A2,,Eton blue
979AAA,,Manatee
987456,,Liver chestnut
98817B,,Cinereous
989898,,Spanish gray
98FF98,,Mint green
9932CC,,Dark orchid
996515,,Golden brown
996666,,Copper rose
9966CC,,Amethyst
997A8D,,Mountbatten pink
9A4EAE,,Purpureus
9AB973,,Olivine
9ACD32,,Yellow-green
9B111E,,Ruby red
9BC4E2,,Pale cerulean
9C2542,,Big dip o’ruby
9C51B6,,Purple Plum
9C7C38,,Metallic Sunburst
9D2933,,Japanese carmine
9E1B32,,Crimson (UA)
9E5E6F,,Rose Dust
9EFD38,,French lime
9F00C5,,Purple (Munsell)
9F00FF,,Vivid violet
9F1D35,,Vivid burgundy
9F2B68,,Amaranth (M&P)
9F4576,,Magenta haze
9F8170,,Beaver
9FA91F,,Citron
A020F0,,Purple (X11)
A020F0,,Veronica
A0D6B4,,Turquoise green
A17A74,,Burnished brown
A1CAF1,,Baby blue eyes
A2006D,,Flirt
A2A2D0,,Blue bell
A2ADD0,,Wild blue yonder
A3C1AD,,Cambridge blue
A45A52,,Redwood
A4DDED,,Non-photo blue
A50021,,University of Pennsylvania red
A50B5E,,Jazzberry jam
A52A2A,,Auburn
A55353,,Middle red purple
A57164,,Blast-off bronze
A57C00,,Gold
A63A79,,Maximum red purple
A67B5B,,Café au lait
A67B5B,,French beige
A67B5B,,Tuscan tan
A6A6A6,,Quick Silver
A75502,,Windsor tan
A7D8DE,,Crystal
A7F432,,Green Lizard
A7FC00,,Spring bud
A81C07,,Rufous
A83731,,Sweet Brown
A8516E,,China rose
A8C3BC,,Opal
A8E4A0,,Granny Smith apple
A91101,,Rojo Spanish red
A99A86,,Grullo
A9B2C3,,Cadet blue (Crayola)
A9BA9D,,Laurel green
AA381E,,Chinese red
AA98A9,,Heliotrope gray
AA98A9,,Rose quartz
AAA9AD,,Silver (Metallic)
AAF0D1,,Magic mint
AB274F,,Amaranth purple
AB4B52,,English red
AB4E52,,Rose vale
AB92B3,,Glossy grape
ACACAC,,Silver chalice
ACACE6,,Maximum blue purple
ACBF60,,Middle green yellow
ACE1AF,,Celadon
ACE5EE,,Blizzard blue
AD4379,,Mystic maroon
AD6F69,,Copper penny
ADD8E6,,Light blue
ADFF2F,,Green-yellow
AE2029,,Upsdell red
AE98AA,,Lilac Luster
AF4035,,Medium carmine
AF6E4D,,Brown sugar
AFDBF5,,Uranian blue
B03060,,Maroon (X11)
B0BF1A,,Acid green
B0C4DE,,Light steel blue
B0E0E6,,Powder blue
B22222,,Firebrick
B284BE,,African violet
B2BEB5,,Ash gray
B2EC5D,,Inchworm
B2FFFF,,Celeste
B2FFFF,,Italian sky blue
B31B1B,,Carnelian
B31B1B,,Cornell red
B3446C,,Irresistible
B3446C,,Raspberry rose
B48395,,English lavender
B53389,,Fandango
B57EDC,,Lavender (floral)
B5B35C,,Olive green
B7410E,,Rust
B768A2,,Pearly purple
B784A7,,Opera mauve
B86D29,,Liver (dogs)
B87333,,Copper
B8860B,,Dark goldenrod
B94E48,,Deep chestnut
B9D9EB,,Columbia Blue
BA160C,,International orange (engineering)
BA55D3,,Medium orchid
BBB477,,Misty moss
BC8F8F,,Rosy brown
BCB88A,,Sage
BCD4E6,,Beau blue
BCD4E6,,Pale aqua
BD33A4,,Byzantine
BDB76B,,Dark khaki
BDDA57,,June bud
BE4F62,,Popstar
BEBEBE,,Gray (X11 gray)
BF00FF,,Electric purple
BF4F51,,Bittersweet shimmer
BFAFB2,,Black Shadows
BFC1C2,,Silver sand
BFFF00,,Bitter lime
BFFF00,,Lime (color wheel)
C0362C,,International orange (Golden Gate Bridge)
C04000,,Mahogany
C0448F,,Red-violet (Crayola)
C08081,,Old rose
C09999,,Tuscany
C0C0C0,,Silver
C0E8D5,,Aero blue
C154C1,,Fuchsia (Crayola)
C19A6B,,Camel
C19A6B,,Desert
C19A6B,,Fallow
C19A6B,,Lion
C19A6B,,Wood brown
C21E56,,Rose red
C2B280,,Ecru
C2B280,,Sand
C30B4E,,Pictorial carmine
C32148,,Bright maroon
C32148,,Maroon (Crayola)
C3B091,,Khaki (web)
C3CDE6,,Periwinkle (Crayola)
C40233,,Red (NCS)
C41E3A,,Cardinal
C46210,,Alloy orange
C4AEAD,,Silver pink
C4C3D0,,Lavender gray
C54B8C,,Mulberry
C5B358,,Vegas gold
C5CBE1,,Light periwinkle
C5E384,,Yellow-green (Crayola)
C71585,,Medium violet-red
C71585,,Red-violet
C72C48,,French raspberry
C74375,,Fuchsia rose
C80815,,Venetian red
C84186,,Smitten
C8509B,,Mulberry (Crayola)
C8A2C8,,Lilac
C8AD7F,,Light French beige
C95A49,,Cedar Chest
C9A0DC,,Wisteria
C9C0BB,,Pale silver
C9C0BB,,Silver (Crayola)
CA1F7B,,Magenta (dye)
CAE00D,,Bitter lemon
CB410B,,Sinopia
CB4154,,Brick red
CB6D51,,Copper red
CBA135,,Satin sheen gold
CC3333,,Persian red
CC3336,,Madder Lake
CC33CC,,Steel pink
CC397B,,Fuchsia purple
CC474B,,English vermillion
CC5500,,Burnt orange
CC7722,,Ochre
CC8899,,Puce
CCA01D,,Lemon curry
CCCCFF,,Lavender blue
CCCCFF,,Periwinkle
CCFF00,,Electric lime
CD5700,,Tenné (tawny)
CD5C5C,,Indian red
CD607E,,Cinnamon Satin
CD7F32,,Bronze
CD9575,,Antique brass
CE2029,,Fire engine red
CE4676,,Ruber
CEFF00,,Volt
CF1020,,Lava
CF3476,,Telemagenta
CF6BA9,,Super pink
CF71AF,,Sky magenta
CFB53B,,Old gold
D0417E,,Magenta (Pantone)
D0F0C0,,Tea green
D0FF14,,Arctic lime
D10047,,Spanish carmine
D10056,,Rubine red
D1E231,,Pear
D2691E,,Chocolate (web)
D2691E,,Cocoa brown
D2B48C,,Tan
D3212D,,Amaranth red
D3AF37,,Metallic gold
D3D3D3,,Light gray
D40000,,Rosso corsa
D44500,,Syracuse Orange
D470A2,,Wild orchid
D473D4,,French mauve
D4AF37,,Gold (metallic)
D65282,,Mystic
D68A59,,Raw sienna
D6CADD,,Languid lavender
D70040,,Carmine (M&P)
D71868,,Dogwood rose
D74894,,Pink (Pantone)
D7837F,,New York pink
D891EF,,Bright lilac
D8B2D1,,Pink lavender
D8BFD8,,Thistle
D9004C,,UA red
D92121,,Maximum red
D9381E,,Vermilion
D982B5,,Middle purple
D98695,,Shimmering Blush
D99058,,Persian orange
D99A6C,,Tan (Crayola)
D9E650,,Maximum green yellow
DA1884,,Barbie Pink
DA2C43,,Rusty red
DA3287,,Deep cerise
DA70D6,,Orchid
DA8A67,,Copper (Crayola)
DA9100,,Harvest gold
DAA520,,Goldenrod
DB2D43,,Alizarin
DBD7D2,,Timberwolf
DC143C,,Crimson
DCDCDC,,Gainsboro
DDA0DD,,Plum (web)
DE3163,,Cerise
DE5285,,Fandango pink
DE5D83,,Blush
DE6FA1,,China pink
DE6FA1,,Liseran purple
DE6FA1,,Thulian pink
DEA5A4,,Pastel pink
DEAA88,,Tumbleweed
DEB887,,Burlywood
DF00FF,,Phlox
DF00FF,,Psychedelic purple
DF73FF,,Heliotrope
DFFF00,,Chartreuse (traditional)
E0115F,,Ruby
E03C31,,CG red
E0B0FF,,Mauve
E0FFFF,,Light cyan
E12C2C,,Permanent Geranium Lake
E1A95F,,Earth yellow
E2062C,,Medium candy apple red
E23D28,,Chili red
E25822,,Flame
E2725B,,Terra cotta
E29CD2,,Orchid (Crayola)
E30022,,Cadmium red
E30B5D,,Raspberry
E3256B,,Razzmatazz
E32636,,Rose madder
E34234,,Cinnabar
E34234,,Vermilion
E3A857,,Indian yellow
E3AB57,,Sunray
E3DAC9,,Bone
E3F988,,Mindaro
E40078,,Red-purple
E4007C,,Mexican pink
E4717A,,Candy pink
E4717A,,Tango pink
E48400,,Fulvous
E49B0F,,Gamboge
E4D00A,,Citrine
E4D96F,,Straw
E52B50,,Amaranth
E58E73,,Middle red
E5AA70,,Fawn
E5E4E2,,Platinum
E60026,,Spanish red
E63E62,,Paradise pink
E68FAC,,Charm pink
E6BE8A,,Gold (Crayola)
E6E6FA,,Lavender (web)
E79FC4,,Kobi
E86100,,Spanish orange
E8CCD7,,Queen pink
E8F48C,,Key lime
E936A7,,Frostbite
E95C4B,,Fire opal
E97451,,Burnt sienna
E9967A,,Dark salmon
E9D66B,,Arylide yellow
E9D66B,,Hansa yellow
E9FFDB,,Nyanza
EAA221,,Marigold
EC5800,,Persimmon
ECB176,,Middle yellow red
ECEBBD,,Pale spring bud
ECEBBD,,Spring green (Crayola)
ED1C24,,Red (pigment)
ED2939,,Imperial red
ED2939,,Red (Pantone)
ED7A9B,,Pale Dogwood
ED7A9B,,Rose Pompadour
ED872D,,Cadmium orange
ED9121,,Carrot orange
EDC9AF,,Desert sand
EDEAE0,,Alabaster
EE204D,,Red (Crayola)
EE82EE,,Violet (web)
EED202,,Safety yellow
EEDC82,,Flax
EEE600,,Titanium yellow
EEED09,,Xanthic
EF98AA,,Mauvelous
EFBBCC,,Cameo pink
EFCC00,,Yellow (Munsell)
EFDECD,,Almond
EFDFBB,,Dutch white
F0599C,,Violet-red(PerBang)
F08080,,Light coral
F0E68C,,Khaki (X11) (Light khaki)
F0E891,,Green-yellow (Crayola)
F0EAD6,,Eggshell
F0F8FF,,Alice blue
F0FFF0,,Honeydew
F0FFFF,,Azure (X11/web color)
F19CBB,,Amaranth pink
F1B42F,,Xanthous
F1DDCF,,Champagne pink
F1E9D2,,Parchment
F2003C,,Red (Munsell)
F28500,,Tangerine
F2BA49,,Maximum yellow red
F2BDCD,,Orchid pink
F2C649,,Maize (Crayola)
F2E8D7,,Magnolia
F37A48,,Mandarin
F38FA9,,Vanilla ice
F3E5AB,,Medium champagne
F3E5AB,,Vanilla
F400A1,,Fashion fuchsia
F400A1,,Hollywood cerise
F4A460,,Sandy brown
F4C2C2,,Baby pink
F4C2C2,,Tea rose
F4C430,,Saffron
F4CA16,,Jonquil
F56FA1,,Cyclamen
F58025,,Princeton orange
F5BD1F,,Orange-yellow
F5DEB3,,Wheat
F5E050,,Minion yellow
F5F5DC,,Beige
F5F5F5,,Cultured
F5FFFA,,Mint cream
F64A8A,,French rose
F653A6,,Magenta (Crayola)
F6ADC6,,Nadeshiko pink
F6EABE,,Lemon meringue
F75394,,Violet-red
F77FBE,,Persian pink
F78FA7,,Pink Sherbet
F7BFBE,,Spanish pink
F7E7CE,,Champagne
F88379,,Congo pink
F88379,,Coral pink
F8B878,,Mellow apricot
F8D568,,Orange-yellow (Crayola)
F8DE7E,,Jasmine
F8DE7E,,Mellow yellow
F8F8FF,,Ghost white
F9429E,,Rose bonbon
FA5053,,Strawberry
FA5B3D,,Orange soda
FA8072,,Salmon
FAD6A5,,Deep champagne
FAD6A5,,Sunset
FAD6A5,,Tuscan
FADA5E,,Naples yellow
FADA5E,,Royal yellow
FADA5E,,Stil de grain yellow
FADADD,,Pale pink
FAE6FA,,Pale purple (Pantone)
FAE7B5,,Banana Mania
FAEBD7,,Antique white
FAF0BE,,Blond
FAF0E6,,Linen
FAFA37,,Maximum yellow
FAFAD2,,Light goldenrod yellow
FB4D46,,Tart Orange
FB607F,,Brink pink
FBAB60,,Rajah
FBCEB1,,Apricot
FBEC5D,,Corn
FBEC5D,,Maize
FC0FC0,,Shocking pink
FC6C85,,Ultra red
FC6C85,,Wild watermelon
FC74FD,,Pink flamingo
FC89AC,,Tickle Me Pink
FCC200,,Golden poppy
FCE883,,Yellow (Crayola)
FCF75E,,Icterine
FD3A4A,,Red Salsa
FD3F92,,French fuchsia
FD6C9E,,French pink
FDBE02,,Mango
FDDDE6,,Piggy pink
FDEE00,,Aureolin
FDF5E6,,Old lace
FDFF00,,Lemon glacier
FE2712,,Red (RYB)
FE28A2,,Persian rose
FE4164,,Neon fuchsia
FE4EDA,,Purple pizzazz
FE6F5E,,Bittersweet
FEBAAD,,Melon
FED8B1,,Light orange
FEDF00,,Yellow (Pantone)
FEFE33,,Yellow (RYB)
FEFEFA,,Baby powder
FF0000,,Red
FF007C,,Winter Sky
FF007F,,Rose
FF0090,,Magenta (process)
FF00FF,,Fuchsia
FF00FF,,Magenta
FF0800,,Candy apple red
FF1493,,Deep pink
FF1DCE,,Hot magenta
FF2400,,Scarlet
FF33CC,,Razzle dazzle rose
FF355E,,Radical Red
FF3800,,Coquelicot
FF3855,,Sizzling Red
FF43A4,,Wild Strawberry
FF4500,,Red-orange (Color wheel)
FF496C,,Infrared
FF4F00,,International orange (aerospace)
FF5349,,Orange-red (Crayola)
FF5349,,Red-orange
FF5470,,Fiery rose
FF55A3,,Brilliant rose
FF5800,,Orange (Pantone)
FF5A36,,Portland Orange
FF6347,,Tomato
FF66CC,,Rose pink
FF6700,,Safety orange (blaze orange)
FF681F,,Orange-red
FF681F,,Red-orange (Crayola)
FF69B4,,Hot pink
FF6E4A,,Outrageous Orange
FF6FFF,,Shocking pink (Crayola)
FF6FFF,,Ultra pink
FF7518,,Pumpkin
FF7538,,Orange (Crayola)
FF7800,,Safety orange
FF7A00,,Heat Wave
FF7E00,,Amber (SAE/ECE)
FF7F00,,Orange
FF7F50,,Coral
FF8243,,Mango Tango
FF8C00,,Dark orange
FF91A4,,Salmon pink
FF91AF,,Baker-Miller pink
FF91AF,,Schauss pink
FF9361,,Strawberry Blonde
FF9505,,Yellow Orange (Color Wheel)
FF9933,,Deep saffron
FF9966,,Atomic tangerine
FF9F00,,Orange peel
FFA07A,,Light salmon
FFA089,,Vivid tangerine
FFA343,,Neon Carrot
FFA500,,Orange (web)
FFA6C9,,Carnation pink
FFA700,,Chrome yellow
FFAA1D,,Bright yellow (Crayola)
FFAE42,,Yellow Orange
FFB200,,Chinese yellow
FFB6C1,,Light pink
FFB7C5,,Cherry blossom pink
FFBA00,,Selective yellow
FFBCD9,,Cotton candy
FFBD88,,Macaroni and Cheese
FFBF00,,Amber
FFC0CB,,Pink
FFC40C,,Mikado yellow
FFC680,,Buff
FFCBA4,,Peach (Crayola)
FFCC33,,Sunglow
FFD300,,Cyber yellow
FFD300,,Yellow (NCS)
FFD700,,Gold (web) (Golden)
FFD800,,School bus yellow
FFDAB9,,Peach puff
FFDAE9,,Mimi pink
FFDB00,,Sizzling Sunrise
FFDB58,,Mustard
FFDDCA,,Unbleached silk
FFDDF4,,Pink lace
FFDEAD,,Navajo white
FFDF00,,Golden yellow
FFE4C4,,Bisque
FFE4E1,,Misty rose
FFE5B4,,Peach
FFEB00,,Middle yellow
FFEBCD,,Blanched almond
FFEF00,,Canary yellow
FFEF00,,Process yellow
FFEF00,,Yellow (process)
FFEFD5,,Papaya whip
FFF0F5,,Lavender blush
FFF44F,,Lemon yellow
FFF5EE,,Seashell
FFF600,,Cadmium yellow
FFF700,,Lemon
FFF700,,Yellow Sunshine
FFF8DC,,Cornsilk
FFF8E7,,Cosmic latte
FFFACD,,Lemon chiffon
FFFAF0,,Floral white
FFFAFA,,Snow
FFFDD0,,Cream
FFFF00,,Yellow
FFFF66,,Laser lemon
FFFF66,,Unmellow yellow
FFFF99,,Canary
FFFF9F,,Lemon yellow (Crayola)
FFFFE0,,Light yellow
FFFFF0,,Ivory
FFFFFF,,White
"""


class RgbFormatter:
    """ Format Color """

    def __init__(self, r=0, g=0, b=0):
        """ 标准的RGB颜色表达方式，数值范围在0~255 """
        self.r, self.g, self.b = r, g, b

    def __repr__(self):
        return f'({self.r}, {self.g}, {self.b})'

    def mixtures(self, rgbs, ratios=1):
        """ 混入另一部分颜色，产生一个新的颜色

        :param rgbs: 新的颜色
        :param ratios: 原颜色权重是1，其他颜色默认权重也是1

        >>> RgbFormatter(255, 0, 0).mixtures([RgbFormatter(0, 255, 0), RgbFormatter(0, 0, 255)])
        (85, 85, 85)
        """
        # 1 非列表结构全部转列表
        if isinstance(rgbs, RgbFormatter):
            rgbs = [rgbs]
        if not isinstance(ratios, (list, tuple)):
            ratios = [ratios]

        # 2 添加当前自身颜色
        rgbs.append(self)
        ratios.append(1)
        # ratios数与rgbs数相同
        if len(ratios) < len(rgbs):
            ratios += [1] * (len(rgbs) - len(ratios))

        # 3 权重计算
        sum_ratio = sum(ratios)
        r = sum([rgb.r * ratio for rgb, ratio in zip(rgbs, ratios)]) / sum_ratio
        g = sum([rgb.g * ratio for rgb, ratio in zip(rgbs, ratios)]) / sum_ratio
        b = sum([rgb.b * ratio for rgb, ratio in zip(rgbs, ratios)]) / sum_ratio
        return RgbFormatter(round(r), round(g), round(b))

    def light(self, ratio=1):
        """ 把颜色变浅

        :param ratio: 原值权重是1，这里可以设置白色的权重
        :return:

        >>> RgbFormatter(0, 0, 0).light()
        (128, 128, 128)
        """
        return self.mixtures(RgbFormatter(255, 255, 255), ratio)

    @staticmethod
    def from_rgb_int255(r=0, g=0, b=0):
        return RgbFormatter(r, g, b)

    def to_tuple(self):
        return self.r, self.g, self.b

    @staticmethod
    def from_vba_value(v):
        """ VBA使用一个Long类型存储颜色，像素(b,g,r) = b*65536 + g*256 + r

        >>> RgbFormatter().from_vba_value(11278293)
        (213, 23, 172)

        # 有时候用宏生成代码，可能会产生负值
        >>> RgbFormatter().from_vba_value(-11260549)
        (123, 45, 84)
        >>> RgbFormatter().from_vba_value(-16776961)
        (255, 0, 0)
        """
        if v < 0: v += 256 ** 3
        rgb = [v % 256, (v // 256) % 256, (v // 65536)]
        return RgbFormatter(*rgb)

    def to_vba_value(self, negative=False):
        """ 转vba的颜色代号值

        :param negative: 默认返回正数，可以设置获得负数表达的数值

        >>> RgbFormatter(123, 45, 84).to_vba_value()
        5516667
        >>> RgbFormatter(123, 45, 84).to_vba_value(negative=True)
        -11260549
        """
        v = self.r + self.g * 256 + self.b * 65536
        if negative: v -= 256 ** 3
        return v

    @property
    def vba_value(self):
        return self.to_vba_value()

    @staticmethod
    def from_hex(s):
        """
        >>> RgbFormatter.from_hex('#7B2D54')
        (123, 45, 84)
        >>> RgbFormatter.from_hex('7B2D54')
        (123, 45, 84)
        """
        m = re.search(r'(\w\w)(\w\w)(\w\w)', s)
        if not m: raise ValueError(f'16进制颜色格式有误: {s}')
        return RgbFormatter(*[int(x, 16) for x in m.groups()])

    def to_hex(self, lower=False):
        """
        :param lower: 默认返回大写，可以设置返回小写

        >>> RgbFormatter(123, 45, 84).to_hex()
        '#7B2D54'
        """
        pattern = '#{:02X}{:02X}{:02X}'
        if lower: pattern = pattern.replace('X', 'x')
        return pattern.format(self.r, self.g, self.b)

    @property
    def hex(self):
        return self.to_hex()

    @staticmethod
    def from_percentage(r, g, b):
        """ 主要用于svg等场景会用的百分率的格式
        输入的每个值是0~1内的百分率浮点数
        """
        return RgbFormatter(*[round(255 * x) for x in (r, g, b)])

    def to_percentage(self):
        """
        >>> RgbFormatter.from_percentage(0.5, 0.7, 0.9).to_percentage()
        [0.5019607843137255, 0.6980392156862745, 0.9019607843137255]
        """
        return [x / 255 for x in (self.r, self.g, self.b)]

    @property
    def percentage(self):
        return self.to_percentage()

    @classmethod
    def from_name(cls, name):
        """ 用名称name来找颜色 """
        _, name2hex = _get_hexs_names(2)
        h = name2hex[name]
        return cls.from_hex(h)

    def distance(self, c2):
        """ 两个rgb颜色的距离对比
        Colour metric: https://www.compuphase.com/cmetric.htm

        我原本想直接欧式距离，但可能不符合人眼的距离感，所以搜了下文档，找了个算法
        """
        rmean = (self.r + c2.r) // 2
        r, g, b = self.r - c2.r, self.g - c2.g, self.b - c2.b
        return math.sqrt((((512 + rmean) * r * r) >> 8) + 4 * g * g + (((767 - rmean) * b * b) >> 8))

    def relative_color_desc(self, relative_color=None, *, color_range=2, precise_mode=False):
        """ 当前self颜色，使用另一个relative_color相对颜色来表达
        一般是任意一个self颜色，采用某个relative_color标准颜色的表示

        :param relative_color: 如果没有输入，则自动找一个最相近的标准颜色

        """
        if relative_color is None:
            relative_color = self.find_similar_std_color(color_range=color_range, precise_mode=precise_mode)

        c0 = relative_color
        h = c0.to_hex()[1:]
        desc = _get_hexs_names(2)[0].get(h, '')
        r, g, b = self.r - c0.r, self.g - c0.g, self.b - c0.b
        r = f'{r:+}' if r else ''
        g = f'{g:+}' if g else ''
        b = f'{b:+}' if b else ''
        return f'({c0.r}{r}, {c0.g}{g}, {c0.b}{b}) {desc}'

    def find_similar_std_color(self, color_range=2, precise_mode=False):
        """
        :param color_range: 色彩搜寻范围，可以写0、1、2
        :param precise_mode:
            False，快速模式，用欧几里得距离计算相似颜色
            True，使用比较好的色彩距离算法
        """

        def fast_func():
            import numpy as np
            color = np.array(self.to_tuple())
            # distances = np.sqrt(np.sum((colors - color) ** 2, axis=1))  # 找最小值，不需要计算sqrt
            distances = np.sum((colors - color) ** 2, axis=1)
            index_of_smallest = np.where(distances == np.amin(distances))
            c = colors[index_of_smallest[0][0]]
            return RgbFormatter(*c.tolist())

        def precise_func():
            """ TODO 这个应该也可以想办法改成np矩阵运算 """
            nearest_color, nearest_dis = None, 1e10
            for c in colors:
                c = RgbFormatter(*c.tolist())
                dis = self.distance(c)
                if dis < nearest_dis:
                    nearest_dis = dis
                    nearest_color = c
            return nearest_color

        colors = _get_colors_array(color_range)
        return precise_func() if precise_mode else fast_func()


@run_once('str')
def _get_hexs_names(color_range=2, zh=True, en=True):
    """ 获得颜色列表

    :param color_range:
        0，标准中文wiki颜色列表
        1，加上扩展的中文颜色
        2，加上所有英文颜色
    :param zh: 是否带上中文名
    :param en: 是否带上英文名

    :return: dict
        key: hex六位编码
        value: 颜色名清单，多个值转为str
    """
    hex2names = defaultdict(list)
    name2hex = {}

    # 1 辅助函数
    def add(line):
        cols = line.split(',')
        if len(cols) < 3:
            return
        h, zh_name, en_name = cols
        if zh and zh_name:
            hex2names[h].append(zh_name)
            name2hex[zh_name] = h
        if en and en_name:
            if en_name.lower() not in hex2names[h]:
                hex2names[h].append(en_name.lower())
            name2hex[en_name] = h
            name2hex[en_name.lower()] = h

    # 2 遍历所有的颜色列表
    for line in _COLOR_LIST0.splitlines():
        add(line)
    if color_range > 0:
        for line in _COLOR_LIST1.splitlines():
            add(line)
    if color_range > 1:
        for line in _COLOR_LIST2.splitlines():
            add(line)

    # 3 names转成str格式
    hex2names = {k: ','.join(v) for k, v in hex2names.items()}
    return hex2names, name2hex


@run_once('str')
def _get_colors_array(color_range=2):
    """ 得到n*3的颜色rgb矩阵

    得到的矩阵，一般是用在后续要计算任意RGB颜色与那个标准色相似
    """
    import numpy as np
    hexs = _get_hexs_names(color_range)[0].keys()
    arr = np.array([RgbFormatter.from_hex(h).to_tuple() for h in hexs])
    return arr



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# @Author : 陈坤泽
# @Email  : 877362867@qq.com
# @Date   : 2022/06/28 21:40

import datetime
import logging
import re
import warnings
import os

import bs4
import pytz
from xml.etree import ElementTree
from tqdm import tqdm

# 过滤这类警告
warnings.filterwarnings("ignore", category=bs4.MarkupResemblesLocatorWarning, module='bs4')

import win32com.client

if win32com.client.gencache.is_readonly:
    win32com.client.gencache.is_readonly = False
    win32com.client.gencache.Rebuild()

from pyxllib.prog.newbie import SingletonForEveryClass
from pyxllib.prog.pupil import Timeout
from pyxllib.file.specialist import XlPath
from pyxllib.algo.treelib import Node

"""
参考了onepy的实现，做了重构。OnePy：Provides pythonic wrappers around OneNote COM interfaces
"""

namespace = "{http://schemas.microsoft.com/office/onenote/2013/onenote}"

# 缓存文件地址
CACHE_DIR = XlPath.tempdir() / 'OneNote/SearchCache'
os.makedirs(CACHE_DIR, exist_ok=True)


class ONProcess(metaclass=SingletonForEveryClass):
    """ onenote 底层win32的接口 """

    def __init__(self, timeout=30):
        """ onenote的win32接口方法是驼峰命名，这个ONProcess做了一层功能封装
        而且估计理论上对所有可以获得的接口都做了封装了

        :param timeout: 读取单个页面的时候，限制用时，单位：秒
            本来只想限制5秒，但发现会有一些页面特别长，需要多一些时间~
            再后来发现还有更慢的页面，半分钟的都有，就再改成30秒了
        """
        # TODO 这里需要针对不同的OneNote版本做自动化兼容，不要让用户填版本
        #   因为让用户填版本，会存在多个实例化对象，使用getxml会有各种问题
        # 目前是支持onenote2016的，但不知道其他版本onenote会怎样
        self.process = win32com.client.gencache.EnsureDispatch('OneNote.Application')
        self.namespace = "{http://schemas.microsoft.com/office/onenote/2013/onenote}"

        # 官方原版的实现，但我觉得可以去掉版本号
        # try:
        #     if version == 16:
        #         self.process = win32com.client.gencache.EnsureDispatch('OneNote.Application')
        #         self.namespace = "{http://schemas.microsoft.com/office/onenote/2013/onenote}"
        #     if version == 15:
        #         self.process = win32com.client.gencache.EnsureDispatch('OneNote.Application.15')
        #         self.namespace = "{http://schemas.microsoft.com/office/onenote/2013/onenote}"
        #     if version == 14:
        #         self.process = win32com.client.gencache.EnsureDispatch('OneNote.Application.14')
        #         self.namespace = "{http://schemas.microsoft.com/office/onenote/2010/onenote}"
        # except Exception as e:
        #     # pywintypes.com_error: (-2147221005, '无效的类字符串', None, None)
        #     # pywintypes.com_error: (-2147221005, '无效的类字符串', None, None)
        #     print(e)
        #     print("error starting onenote {}".format(version))

        global namespace
        namespace = self.namespace

        self.timeout_seconds = timeout

    def get_hierarchy(self, start_node_id="", hierarchy_scope=4):
        """
          HierarchyScope
          0 - Gets just the start node specified and no descendants.
          1 - Gets the immediate child nodes of the start node, and no descendants in higher or lower subsection groups.
          2 - Gets all notebooks below the start node, or root.
          3 - Gets all sections below the start node, including sections in section groups and subsection groups.
          4 - Gets all pages below the start node, including all pages in section groups and subsection groups.
        """
        return self.process.GetHierarchy(start_node_id, hierarchy_scope)

    def update_hierarchy(self, changes_xml_in):
        try:
            self.process.UpdateHierarchy(changes_xml_in)
        except Exception as e:
            print("Could not Update Hierarchy")

    def open_hierarchy(self, path, relative_to_object_id, object_id, create_file_type=0):
        """
          CreateFileType
          0 - Creates no new object.
          1 - Creates a notebook with the specified name at the specified location.
          2 - Creates a section group with the specified name at the specified location.
          3 - Creates a section with the specified name at the specified location.
        """
        try:
            return self.process.OpenHierarchy(path, relative_to_object_id, "", create_file_type)
        except Exception as e:
            print("Could not Open Hierarchy")

    def delete_hierarchy(self, object_id, excpect_last_modified=""):
        try:
            self.process.DeleteHierarchy(object_id, excpect_last_modified)
        except Exception as e:
            print("Could not Delete Hierarchy")

    def create_new_page(self, section_id, new_page_style=0):
        """
          NewPageStyle
          0 - Create a Page that has Default Page Style
          1 - Create a blank page with no title
          2 - Createa blank page that has no title
        """
        try:
            self.process.CreateNewPage(section_id, "", new_page_style)
        except Exception as e:
            print("Unable to create the page")

    def close_notebook(self, notebook_id):
        try:
            self.process.CloseNotebook(notebook_id)
        except Exception as e:
            print("Could not Close Notebook")

    def get_page_content(self, page_id, page_info=0):
        """
          PageInfo
          0 - Returns only file page content, without selection markup and binary data objects. This is the standard value to pass.
          1 - Returns page content with no selection markup, but with all binary data.
          2 - Returns page content with selection markup, but no binary data.
          3 - Returns page content with selection markup and all binary data.
        """
        with Timeout(self.timeout_seconds):
            return self.process.GetPageContent(page_id, "", page_info)

    def update_page_content(self, page_changes_xml_in, excpect_last_modified=0):
        try:
            self.process.UpdatePageContent(page_changes_xml_in, excpect_last_modified)
        except Exception as e:
            print("Could not Update Page Content")

    def get_binary_page_content(self, page_id, callback_id):
        try:
            return self.process.GetBinaryPageContent(page_id, callback_id)
        except Exception as e:
            print("Could not Get Binary Page Content")

    def delete_page_content(self, page_id, object_id, excpect_last_modified=0):
        try:
            self.process.DeletePageContent(page_id, object_id, excpect_last_modified)
        except Exception as e:
            print("Could not Delete Page Content")

    # Actions

    def navigate_to(self, object_id, new_window=False):
        try:
            self.process.NavigateTo(object_id, "", new_window)
        except Exception as e:
            print("Could not Navigate To")

    def publish(self, hierarchy_id, target_file_path, publish_format, clsid_of_exporter=""):
        """
         PublishFormat
          0 - Published page is in .one format.
          1 - Published page is in .onea format.
          2 - Published page is in .mht format.
          3 - Published page is in .pdf format.
          4 - Published page is in .xps format.
          5 - Published page is in .doc or .docx format.
          6 - Published page is in enhanced metafile (.emf) format.
        """
        try:
            self.process.Publish(hierarchy_id, target_file_path, publish_format, clsid_of_exporter)
        except Exception as e:
            print("Could not Publish")

    def open_package(self, path_package, path_dest):
        try:
            return self.process.OpenPackage(path_package, path_dest)
        except Exception as e:
            print("Could not Open Package")

    def get_hyperlink_to_object(self, hierarchy_id, target_file_path=""):
        try:
            return self.process.GetHyperlinkToObject(hierarchy_id, target_file_path)
        except Exception as e:
            print("Could not Get Hyperlink")

    def find_pages(self, start_node_id, search_string, display):
        try:
            return self.process.FindPages(start_node_id, search_string, "", False, display)
        except Exception as e:
            print("Could not Find Pages")

    def get_special_location(self, special_location=0):
        """
          SpecialLocation
          0 - Gets the path to the Backup Folders folder location.
          1 - Gets the path to the Unfiled Notes folder location.
          2 - Gets the path to the Default Notebook folder location.
        """
        try:
            return self.process.GetSpecialLocation(special_location)
        except Exception as e:
            print("Could not retreive special location")


class _CommonMethods:
    """ 笔记本、分区组、分区、页面 共有的一些成员方法 """

    @property
    def ancestors(self):
        """ 获得所有父结点 """
        parents = []
        p = self
        while getattr(p, 'parent', False):
            parents.append(p.parent)
            p = p.parent
        return reversed(parents)

    @property
    def abspath_name(self):
        names = [x.name for x in self.ancestors]
        names.append(self.name)
        return '/'.join(names)

    def search(self, pattern, child_depth=1, *, return_mode='text', padding_mode=0):
        """ 查找内容

        Page、Section、SectionGroup等实际所用的search方法

        :param pattern:
            text, 检索出现该关键词的node，并展示其相关的上下文内容
            func, 可以输入自定义函数 check_node(node)->bool，True表示符合检索条件的node
            re.compile，可以输入编译的正则模式，会使用re.search进行匹配
        :param int child_depth: 对于检索到的node，向下展开几层子结点
            -1，表示全部展开
            0，表示不展开子结点
            1，只展开直接子结点
            ...
        :param return_mode:
            text，文本展示
            html，网页富文本
        """
        # 1 按照规则检索内容
        if isinstance(pattern, str):
            def check_node(node):
                # 纯文本部分
                if pattern in node.name:
                    return True
                # 如果纯文本找不到，也会在富文本格式里尝试匹配
                html_text = getattr(node, '_html_content', '')
                if html_text and pattern in html_text:
                    return True
                return False

        elif isinstance(pattern, re.Pattern):
            def check_node(node):
                if pattern.search(node.name):
                    return True
                html_text = getattr(node, '_html_content', '')
                if html_text and pattern.search(html_text):
                    return True
                return False
        else:
            check_node = pattern

        root = self._search(print_mode=True)  # 显示进度条
        root.sign_node(check_node, flag_name='_flag', child_depth=child_depth)

        # 2 展示内容
        if return_mode == 'text':
            return root.render(filter_=lambda x: getattr(x, '_flag', 0))
        elif return_mode == 'html':
            return root.render_html('_html_content',
                                    filter_=lambda x: getattr(x, '_flag', 0),
                                    padding_mode=padding_mode)
        else:
            raise ValueError


class OneNote(ONProcess, _CommonMethods):
    """ OneNote软件，这是一个单例类

    注意，从ONProcess继承的OneNote也是单例类
    但从ONProcess、OneNote生成的是不同的两个对象
    """

    def __init__(self, timeout=30):
        """
        如果出现这个错误：This COM object can not automate the makepy process - please run makepy manually for this object
            可以照 https://github.com/varunsrin/one-py 文章末尾的方式操作
            把 HKEY_CLASSES_ROOT\TypeLib\{0EA692EE-BB50-4E3C-AEF0-356D91732725} 的 1.0 删掉
            （这个 KEY ID 值大家电脑上都是一样的）
        """
        # trick: 这里有跟单例类有关的一些问题，导致ONProcess需要提前初始化一次
        super().__init__(timeout)

        self.xml = self.get_hierarchy("", 4)
        self.object_tree = ElementTree.fromstring(self.xml)
        self.hierarchy = Hierarchy(self.object_tree)
        self._children = list(self.hierarchy)
        self.name = 'onenote'

    def get_page_content(self, page_id):
        page_content_xml = ElementTree.fromstring(super(OneNote, self).get_page_content(page_id))
        return PageContent(page_content_xml)

    def update_page_content(self, page_changes_xml_in):
        """
        :param page_changes_xml_in:
            xml，可以是原始的xml文本
                onenote.update_page_content(page.getxml().replace('曹一众', '曹二众'))
            soup, 可以传入一个bs4的soup对象

        这里设置pytz时间的东西我也看不懂，但大受震撼~~有点莫名其妙
        How to debug win32com call in python：
        https://stackoverflow.com/questions/34904094/how-to-debug-win32com-call-in-python/34979646#34979646
        """
        return super(OneNote, self).update_page_content(page_changes_xml_in,
                                                        pytz.utc.localize(datetime.datetime(1899, 12, 30)))

    def names(self):
        """ 所有笔记本的名称 """
        ls = list(map(lambda x: x.name, self.hierarchy))
        return ls

    def nicknames(self):
        """ 所有笔记本的昵称 """
        ls = list(map(lambda x: x.nickname, self.hierarchy))
        return ls

    def __getitem__(self, item):
        """ 通过编号或名称索引获得笔记本 """
        return self.hierarchy[item]

    def _search(self, *, print_mode=False):
        root = Node('Notebook：' + self.name)
        for x in tqdm(self._children, desc='解析笔记本中', smoothing=0, disable=not print_mode):
            cur_node = x._search()
            cur_node.parent = root

        return root


class Hierarchy:

    def __init__(self, xml=None):
        self.xml = xml
        self._children = []
        if xml is not None:
            self.__deserialize_from_xml(xml)

    def __deserialize_from_xml(self, xml):
        self._children = [Notebook(n) for n in xml]

    def __iter__(self):
        for c in self._children:
            yield c

    def __getitem__(self, item):
        """通过编号或名称索引子节点内容"""
        if isinstance(item, int):
            return self._children[item]
        elif isinstance(item, str):
            for nb in self:
                if nb.nickname == item:
                    return nb
        return None


class HierarchyNode:

    def __init__(self, parent=None):
        self.name = ""
        self.path = ""
        self.id = ""
        self.last_modified_time = ""
        self.synchronized = ""

    def deserialize_from_xml(self, xml):
        self.name = xml.get("name")
        self.path = xml.get("path")
        self.id = xml.get("ID")
        self.last_modified_time = xml.get("lastModifiedTime")


class Notebook(HierarchyNode, _CommonMethods):

    def __init__(self, xml=None):
        self.xml = xml
        super().__init__(self)
        self.nickname = ""
        self.color = ""
        self.is_currently_viewed = ""
        self.recycleBin = None
        self._children = []
        if xml is not None:
            self.__deserialize_from_xml(xml)

    def __deserialize_from_xml(self, xml):
        HierarchyNode.deserialize_from_xml(self, xml)
        self.nickname = xml.get("nickname")
        self.color = xml.get("color")
        self.is_currently_viewed = xml.get("isCurrentlyViewed")
        self.recycleBin = None
        for node in xml:
            if node.tag == namespace + "Section":
                self._children.append(Section(node, self))

            elif node.tag == namespace + "SectionGroup":
                if node.get("isRecycleBin"):
                    self.recycleBin = SectionGroup(node, self)
                else:
                    self._children.append(SectionGroup(node, self))

    def __iter__(self):
        for c in self._children:
            yield c

    def __str__(self):
        return self.name

    def __getitem__(self, item):
        """通过编号或名称索引子节点内容"""
        if isinstance(item, int):
            return self._children[item]
        elif isinstance(item, str):
            for nb in self:
                if nb.name == item:
                    return nb
        return None

    def _search(self, *, print_mode=False):
        root = Node('Notebook：' + self.name)
        for x in tqdm(self._children, desc='解析分区(组)中', smoothing=0, disable=not print_mode):
            cur_node = x._search()
            cur_node.parent = root

        return root


class SectionGroup(HierarchyNode, _CommonMethods):
    """ 分区组 """

    def __init__(self, xml=None, parent_node=None):
        self.xml = xml
        super().__init__(self)
        self.is_recycle_Bin = False
        self._children = []
        self.parent = parent_node
        if xml is not None:
            self.__deserialize_from_xml(xml)

    def __iter__(self):
        # ckz: 这个遍历的时候，就是OneNote里看到的从左到右的顺序：先所有分区，然后所有分区组
        #
        for c in self._children:
            yield c

    def __str__(self):
        return self.name

    def __deserialize_from_xml(self, xml):
        HierarchyNode.deserialize_from_xml(self, xml)
        self.is_recycle_Bin = xml.get("isRecycleBin")
        for node in xml:
            if node.tag == namespace + "SectionGroup":
                self._children.append(SectionGroup(node, self))
            if node.tag == namespace + "Section":
                self._children.append(Section(node, self))

    def __getitem__(self, item):
        """ 通过 编号 或 名称 索引子节点内容

        注意使用字符串引用的时候，可能会有重名的问题！
        """
        if isinstance(item, int):
            return self._children[item]
        elif isinstance(item, str):
            for nb in self:
                if nb.name == item:
                    return nb
        return None

    def get_page_num(self):
        return sum([x.get_page_num() for x in self._children])

    def _search(self, *, print_mode=False):
        root = Node('SectionGroup：' + self.name)
        for x in tqdm(self._children, desc='解析分区(组)中', smoothing=0, disable=not print_mode):
            cur_node = x._search()
            cur_node.parent = root
        return root


class Section(HierarchyNode, _CommonMethods):
    """ 分区 """

    def __init__(self, xml=None, parent_node=None):
        self.xml = xml
        super().__init__(self)
        self.color = ""
        self.read_only = False
        self.is_currently_viewed = False
        self._children = []
        self.parent = parent_node
        if xml is not None:
            self.__deserialize_from_xml(xml)

    def __iter__(self):
        for c in self._children:
            yield c

    def __str__(self):
        return self.name

    def __deserialize_from_xml(self, xml):
        HierarchyNode.deserialize_from_xml(self, xml)
        self.color = xml.get("color")
        try:
            self.read_only = xml.get("readOnly")
        except Exception as e:
            self.read_only = False
        try:
            self.is_currently_viewed = xml.get("isCurrentlyViewed")
        except Exception as e:
            self.is_currently_viewed = False

        self._children = [Page(node, self) for node in xml]

    def __getitem__(self, item):
        """通过编号或名称索引子节点内容"""
        if isinstance(item, int):
            return self._children[item]
        elif isinstance(item, str):
            for nb in self:
                if nb.name == item:
                    return nb
        return None

    def get_page_num(self):
        return len(self._children)

    def _search(self, *, print_mode=False):
        root = Node('Section：' + self.name)
        page_lv1, page_lv2 = root, root

        for x in tqdm(self._children, desc='解析页面中', smoothing=0, disable=not print_mode):
            # print(x.name)
            cur_page = x._search()
            if x.page_level == '1':
                cur_page.parent = root
                page_lv1 = cur_page
            elif x.page_level == '2':
                cur_page.parent = page_lv1
                page_lv2 = cur_page
            else:
                cur_page.parent = page_lv2

        return root


class Page(_CommonMethods):
    """ 页面 """

    def __init__(self, xml=None, parent_node=None):
        self.xml = xml
        self.name = ""
        self.id = ""
        self.date_time = ""
        self.last_modified_time = ""
        self.page_level = ""
        self.is_currently_viewed = ""
        self._children = []
        self.parent = parent_node
        if xml is not None:  # != None is required here, since this can return false
            self.__deserialize_from_xml(xml)

    def __iter__(self):
        for c in self._children:
            yield c

    def __str__(self):
        return self.name

        # Get / Set Meta

    @property
    def root(self):
        p = self
        while getattr(p, 'parent', False):
            p = p.parent
        return p

    def __deserialize_from_xml(self, xml):
        self.xml = xml
        self.name = xml.get("name")
        self.id = xml.get("ID")
        self.date_time = xml.get("dateTime")
        self.last_modified_time = xml.get("lastModifiedTime")
        self.page_level = xml.get("pageLevel")
        self.is_currently_viewed = xml.get("isCurrentlyViewed")
        self._children = [Meta(node) for node in xml]

    def getxml(self, page_info=0):
        """ 获得页面的xml内容 """
        try:
            res = super(OneNote, onenote).get_page_content(self.id, page_info)
        except TimeoutError as e:
            e.args = [e.args[0] + f'\n\t{self.abspath_name} 页面获取失败，请检查可能包含的office公式并删除。' \
                                  f'并且看下您的OneNote可能无响应了，请重启OneNote。']
            raise e

        if res is None:
            logging.warning(f'{self.abspath_name} 未成功提取页面内容')

        return res

    def browser_xml(self):
        from pyxllib.debug.specialist import browser, XlPath
        xml = self.getxml()
        browser(xml, file=XlPath.tempfile('.xml'))

    def parse_oe_tree(self, root=None):
        """ 获得本Page页面的树形结构，返回一个Node根节点

        因为层次结构有多种不同的表现形式
            层次结构A：Outline 文本框
            层次结构B：h1、h2、h3 等标题结构
            层次结构C：正文里的缩进层级
        所以实现这里parent的引用，算法会稍复杂。需要动态更新，实时返回新的parent。
            比如在遍历tag.contents的时候，因为[层次结构C]的原因，会出现 parent = dfs_parse_node(y, parent) 的较奇怪的写法
            在parse_oe中，parent的层次实现规则，也会较复杂，有些trick
        """
        from pyxllib.text.xmllib import BeautifulSoup

        xml = self.getxml()
        soup = BeautifulSoup(xml or '', 'xml')

        # self.browser_xml()  # 可以用这个查原始的xml内容

        def parse_oe(x, parent):
            # 1 获得3个主要属性
            def get_text(x):
                if y := x.find('T', recursive=False):
                    t1 = BeautifulSoup(y.text, 'lxml').text
                    t2 = y.text
                elif y := x.find('Table', recursive=False):
                    # 先Columns标记了一共m列，每列的宽度
                    # 然后每一行是一个Row，里面有m个Cell
                    t1 = '[Table]'
                    t2 = t1
                elif y := x.find('Image', recursive=False):
                    t1 = '[Image]'
                    t2 = t1
                else:
                    t1 = ''
                    t2 = ''
                return t1, t2

            style_name = style_defs.get(x.get('quickStyleIndex', ''), '')
            pure_text, html_text = get_text(x)  # 文本内容
            m = x.find('OEChildren', recursive=False)  # 文本性质子结点

            # 空数据跳过
            # if not pure_text and not m:
            #     return parent

            # 2 处理层次结构B
            if re.match(r'h\d$', style_name):  # 标题类
                while True:
                    parent_style_name = getattr(parent, '_style_name', '')
                    if re.match(r'h\d$', parent_style_name) and parent_style_name >= style_name:
                        # 如果父结点也是标题类型，且数值上不大于当前结点，则当前结点的实际父结点要往上层找
                        parent = parent.parent
                    else:
                        break
                # 标题类，会重置parent，本身作为一个中间结点
                cur_node = parent = Node(pure_text, parent, _style_name=style_name, _html_content=html_text)
            else:
                cur_node = Node(pure_text, parent, _html_content=html_text)

            # 3 表格、图片等特殊结构增设层级
            if pure_text.startswith('[Table]'):
                for z in x.find_all('T'):
                    Node(BeautifulSoup(z.text, 'lxml').text, cur_node, _html_content=z.text)
            elif pure_text.startswith('[Image]'):
                y = x.find('Image', recursive=False)
                for z in y.get('alt', '').splitlines():
                    Node(z, cur_node)

            # 4 处理层次结构C
            if m:
                for y in m.find_all('OE', recursive=False):
                    parse_oe(y, cur_node)

            return parent

        def dfs_parse_node(x, parent):
            if isinstance(x, bs4.element.Tag):
                if x.name == 'QuickStyleDef':
                    style_defs[x['index']] = x['name']
                elif x.name == 'Title':
                    parent.root.name = BeautifulSoup(x.T.text, 'lxml').text
                elif x.name == 'Outline':
                    # 处理层次结构A
                    nonlocal outline_cnt
                    if outline_cnt:
                        if outline_cnt == 1:
                            parent = Node(f'Outline{outline_cnt}', parent)
                        else:
                            parent = Node(f'Outline{outline_cnt}', parent.parent)
                        outline_cnt += 1
                    for y in x.contents:
                        parent = dfs_parse_node(y, parent)
                elif x.name == 'OE':
                    parent = parse_oe(x, parent)
                else:
                    for y in x.contents:
                        parent = dfs_parse_node(y, parent)
            return parent  # parent有可能会更新

        if root is None:
            root = Node('root')
        style_defs = {}
        # trick: outline_cnt不仅用来标记是否有多个Outline需要设中介结点。也记录了当前Outline的编号。
        outline_cnt = 1 if len(soup.find_all('Outline')) > 1 else 0
        dfs_parse_node(soup, root)
        return root

    def get_page_num(self):
        return 1

    def _search(self, *, print_mode=False):
        """ 先生成所有结点

        :param print_mode: 统一所有的_search定义范式，虽然在Page这里其实没用~

        """
        file = CACHE_DIR / (self.id + self.last_modified_time.replace(':', '') + '.pkl')

        if file.is_file() and False:
            root = file.read_pkl()
        else:
            root = self.parse_oe_tree()
            root.name = self.name

            if not root.is_leaf:
                # 删除旧时间点的缓存文件，存储新的缓存文件
                for f in CACHE_DIR.glob(f'{self.id}*.pkl'):
                    f.delete()
                file.write_pkl(root)

        return root


class Meta:

    def __init__(self, xml=None):
        self.xml = xml
        self.name = ""
        self.content = ""
        if xml is not None:
            self.__deserialize_from_xml(xml)

    def __str__(self):
        return self.name

    def __deserialize_from_xml(self, xml):
        self.name = xml.get("name")
        self.id = xml.get("content")

    def getxml(self):
        """获得页面的"""
        return super(OneNote, onenote).get_page_content(self.id)


class PageContent:

    def __init__(self, xml=None):
        self.xml = xml
        self.name = ""
        self.id = ""
        self.date_time = ""
        self.last_modified_time = ""
        self.page_level = ""
        self.lang = ""
        self.is_currently_viewed = ""
        self._children = []
        self.files = []
        if xml is not None:
            self.__deserialize_from_xml(xml)

    def __iter__(self):
        for c in self._children:
            yield c

    def __str__(self):
        return self.name

    def __deserialize_from_xml(self, xml):
        self.name = xml.get("name")
        self.id = xml.get("ID")
        self.date_time = xml.get("dateTime")
        self.last_modified_time = xml.get("lastModifiedTime")
        self.page_level = xml.get("pageLevel")
        self.lang = xml.get("lang")
        self.is_currently_viewed = xml.get("isCurrentlyViewed")
        for node in xml:
            if node.tag == namespace + "Outline":
                self._children.append(Outline(node))
            elif node.tag == namespace + "Ink":
                self.files.append(Ink(node))
            elif node.tag == namespace + "Image":
                self.files.append(Image(node))
            elif node.tag == namespace + "InsertedFile":
                self.files.append(InsertedFile(node))
            elif node.tag == namespace + "Title":
                self._children.append(Title(node))


class Title:

    def __init__(self, xml=None):
        self.xml = xml
        self.style = ""
        self.lang = ""
        self._children = []
        if xml is not None:
            self.__deserialize_from_xml(xml)

    def __str__(self):
        return "Page Title"

    def __iter__(self):
        for c in self._children:
            yield c

    def __deserialize_from_xml(self, xml):
        self.style = xml.get("style")
        self.lang = xml.get("lang")
        for node in xml:
            if node.tag == namespace + "OE":
                self._children.append(OE(node, self))


class Outline:

    def __init__(self, xml=None):
        self.xml = xml
        self.author = ""
        self.author_initials = ""
        self.last_modified_by = ""
        self.last_modified_by_initials = ""
        self.last_modified_time = ""
        self.id = ""
        self._children = []
        if xml is not None:
            self.__deserialize_from_xml(xml)

    def __iter__(self):
        for c in self._children:
            yield c

    def __str__(self):
        return "Outline"

    def __deserialize_from_xml(self, xml):
        self.author = xml.get("author")
        self.author_initials = xml.get("authorInitials")
        self.last_modified_by = xml.get("lastModifiedBy")
        self.last_modified_by_initials = xml.get("lastModifiedByInitials")
        self.last_modified_time = xml.get("lastModifiedTime")
        self.id = xml.get("objectID")
        append = self._children.append
        for node in xml:
            if node.tag == namespace + "OEChildren":
                for childNode in node:
                    if childNode.tag == namespace + "OE":
                        append(OE(childNode, self))


class Position:

    def __init__(self, xml=None, parent_node=None):
        self.xml = xml
        self.x = ""
        self.y = ""
        self.z = ""
        self.parent = parent_node
        if xml is not None:
            self.__deserialize_from_xml(xml)

    def __deserialize_from_xml(self, xml):
        self.x = xml.get("x")
        self.y = xml.get("y")
        self.z = xml.get("z")


class Size:

    def __init__(self, xml=None, parent_node=None):
        self.xml = xml
        self.width = ""
        self.height = ""
        self.parent = parent_node
        if xml is not None:
            self.__deserialize_from_xml(xml)

    def __deserialize_from_xml(self, xml):
        self.width = xml.get("width")
        self.height = xml.get("height")


class OE:

    def __init__(self, xml=None, parent_node=None):
        self.xml = xml
        self.creation_time = ""
        self.last_modified_time = ""
        self.last_modified_by = ""
        self.id = ""
        self.alignment = ""
        self.quick_style_index = ""
        self.style = ""
        self.text = ""
        self._children = []
        self.parent = parent_node
        self.files = []
        if xml is not None:
            self.__deserialize_from_xml(xml)

    def __iter__(self):
        for c in self._children:
            yield c

    def __str__(self):
        try:
            return self.text
        except AttributeError:
            return "Empty OE"

    def __deserialize_from_xml(self, xml):
        self.creation_time = xml.get("creationTime")
        self.last_modified_time = xml.get("lastModifiedTime")
        self.last_modified_by = xml.get("lastModifiedBy")
        self.id = xml.get("objectID")
        self.alignment = xml.get("alignment")
        self.quick_style_index = xml.get("quickStyleIndex")
        self.style = xml.get("style")

        for node in xml:
            if node.tag == namespace + "T":
                if node.text is not None:
                    self.text = node.text
                else:
                    self.text = "NO TEXT"

            elif node.tag == namespace + "OEChildren":
                for childNode in node:
                    if childNode.tag == namespace + "OE":
                        self._children.append(OE(childNode, self))

            elif node.tag == namespace + "Image":
                self.files.append(Image(node, self))

            elif node.tag == namespace + "InkWord":
                self.files.append(Ink(node, self))

            elif node.tag == namespace + "InsertedFile":
                self.files.append(InsertedFile(node, self))


class InsertedFile:

    # need to add position data to this class

    def __init__(self, xml=None, parent_node=None):
        self.xml = xml
        self.path_cache = ""
        self.path_source = ""
        self.preferred_name = ""
        self.last_modified_time = ""
        self.last_modified_by = ""
        self.id = ""
        self.parent = parent_node
        if xml is not None:
            self.__deserialize_from_xml(xml)

    def __iter__(self):
        yield None

    def __str__(self):
        try:
            return self.preferredName
        except AttributeError:
            return "Unnamed File"

    def __deserialize_from_xml(self, xml):
        self.path_cache = xml.get("pathCache")
        self.path_source = xml.get("pathSource")
        self.preferred_name = xml.get("preferredName")
        self.last_modified_time = xml.get("lastModifiedTime")
        self.last_modified_by = xml.get("lastModifiedBy")
        self.id = xml.get("objectID")


class Ink:

    # need to add position data to this class

    def __init__(self, xml=None, parent_node=None):
        self.xml = xml
        self.recognized_text = ""
        self.x = ""
        self.y = ""
        self.ink_origin_x = ""
        self.ink_origin_y = ""
        self.width = ""
        self.height = ""
        self.data = ""
        self.callback_id = ""
        self.parent = parent_node

        if xml is not None:
            self.__deserialize_from_xml(xml)

    def __iter__(self):
        yield None

    def __str__(self):
        try:
            return self.recognizedText
        except AttributeError:
            return "Unrecognized Ink"

    def __deserialize_from_xml(self, xml):
        self.recognized_text = xml.get("recognizedText")
        self.x = xml.get("x")
        self.y = xml.get("y")
        self.ink_origin_x = xml.get("inkOriginX")
        self.ink_origin_y = xml.get("inkOriginY")
        self.width = xml.get("width")
        self.height = xml.get("height")

        for node in xml:
            if node.tag == namespace + "CallbackID":
                self.callback_id = node.get("callbackID")
            elif node.tag == namespace + "Data":
                self.data = node.text


class Image:

    def __init__(self, xml=None, parent_node=None):
        self.xml = xml
        self.format = ""
        self.original_page_number = ""
        self.last_modified_time = ""
        self.id = ""
        self.callback_id = None
        self.data = ""
        self.parent = parent_node
        if xml is not None:
            self.__deserialize_from_xml(xml)

    def __iter__(self):
        yield None

    def __str__(self):
        return self.format + " Image"

    def __deserialize_from_xml(self, xml):
        self.format = xml.get("format")
        self.original_page_number = xml.get("originalPageNumber")
        self.last_modified_time = xml.get("lastModifiedTime")
        self.id = xml.get("objectID")
        for node in xml:
            if node.tag == namespace + "CallbackID":
                self.callback_id = node.get("callbackID")
            elif node.tag == namespace + "Data":
                if node.text is not None:
                    self.data = node.text


onenote = OneNote()
onenote2 = OneNote()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# @Author : 陈坤泽
# @Email  : 877362867@qq.com
# @Date   : 2022/04/07 15:45

"""
树形结构相关的处理
"""

from pyxllib.prog.pupil import check_install_package

check_install_package('anytree')

import re

from anytree import Node

from pyxllib.prog.pupil import EnchantBase, run_once


class EnchantNode(EnchantBase):
    @classmethod
    @run_once
    def enchant(cls):
        names = cls.check_enchant_names([Node])
        cls._enchant(Node, names)

    @staticmethod
    def sign_node(self, check_func, *, flag_name='_flag', child_depth=-1):
        """ 遍历所有结点

        :param check_func: 检查node是否符合要求
            True，自身打上标记2，其余父节点、子结点全部打上1 （这里处理方式后续可以再扩展更灵活）
        """
        from anytree import PreOrderIter

        if child_depth == -1:
            child_depth = None
        else:
            child_depth += 1

        def set_flag(node, flag):
            # 结点没有值或者flag小余当前值时，才标记更新为当前flag
            setattr(node, flag_name, max(flag, getattr(node, flag_name, 0)))

        for x in PreOrderIter(self, filter_=check_func):
            # 对找到的结点标记2
            set_flag(x, 2)
            # 其父结点、子结点标记1
            for y in x.ancestors:
                set_flag(y, 1)
            for y in PreOrderIter(x, maxlevel=child_depth):
                set_flag(y, 1)

    @staticmethod
    def render(self, *, childiter=list, maxlevel=None, filter_=None, dedent=0):
        """

        也可以这样可视化：
        >> from anytree import RenderTree
        >> print(RenderTree(root))
        """
        from anytree import RenderTree

        def fold_text(s):
            return s.replace('\n', ' ')[:150]

        ls = []
        for pre, fill, node in RenderTree(self, childiter=childiter, maxlevel=maxlevel):
            if filter_:
                if not filter_(node):
                    continue
                # 使用filter_参数时，前缀统一成空格
                pre = (len(pre) // 4 - dedent) * '\t'

            msg = re.search(r',\s*(.+?)\)', str(node))  # 显示node的属性值
            msg = ('\t' + msg.group(1)) if msg else ''
            ls.append(fold_text(f'{pre}{node.name}{msg}'))

        return '\n'.join(ls)

    @staticmethod
    def render_html(self, html_attr_name='name', *,
                    childiter=list, maxlevel=None, filter_=None, padding_mode=0,
                    dedent=0):
        """ 渲染成html页面内容

        :param html_attr_name: 最好原node有一个字段存储需要渲染的html内容
            默认使用name的值作为渲染对象
        :param padding_mode:
            0，默认配置，使用html自带的悬挂缩进功能。这种展示效果最好，但复制内容的时候不带缩进。
            1，使用空格缩进。长文本折行显示效果不好。但是方便复制。
        :param dedent: 统一减少缩进层级

        也可以这样可视化：
        >> from anytree import RenderTree
        >> print(RenderTree(root))
        """
        from anytree import RenderTree

        ls = []
        for pre, fill, node in RenderTree(self, childiter=childiter, maxlevel=maxlevel):
            if filter_ and not filter_(node):
                continue
            pre_len = len(pre) - dedent * 4

            msg = re.search(r',\s*(.+?)\)', str(node))  # 显示node的属性值
            msg = ('\t' + msg.group(1)) if msg else ''
            content = f'{getattr(node, html_attr_name, node.name)}{msg}'
            if not content.strip():
                content = '<br/>'
            if padding_mode == 1:
                ls.append(f'<div>{"&nbsp;" * pre_len}{content}</div>')
            else:
                ls.append(f'<div style="padding-left:{pre_len // 2 + 1}em;text-indent:-1em">{content}</div>')

        return '\n'.join(ls)


EnchantNode.enchant()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# @Author : 陈坤泽
# @Email  : 877362867@qq.com
# @Date   : 2022/04/12 08:59

import json
import sqlite3

import pandas as pd


class SqlBase:
    """ Sql语法通用的功能 """

    def __改(self):
        pass

    def create_index(self, index_name, table_name, cols):
        if not isinstance(cols, str):
            cols = ','.join(map(str, cols))
        self.execute(f'CREATE INDEX {index_name} ON {table_name}(cols)')

    def __查(self):
        pass

    def select_col(self, table_name, col):
        """ 【查】获取一个表的一列数据

        这个功能其实可以封装在更底层的Cursor，类似fetchone、fetchall的接口，但每次获取会自动去掉外层的tuple
        """
        for x in self.execute(f'SELECT {col} FROM {table_name}'):
            yield x[0]

    def group_count(self, table_name, cols, count_column_name='cnt'):
        """ 【查】分组统计各组值组合出现次数
        分析{table}表中，{cols}出现的种类和次数，按照出现次数从多到少排序

        :param str|list cols: 输入逗号','隔开的字符串，比如
            con.group_count('gpu_trace', 'host_name,total_memory')
            后记：写list也行，会自动join为字符串
        """
        if not isinstance(cols, str):
            cols = ','.join(map(str, cols))
        sql = f'SELECT {cols}, COUNT(*) {count_column_name} FROM {table_name} ' \
              f'GROUP BY {cols} ORDER BY {count_column_name} DESC'
        records = self.execute(sql).fetchall()
        df = pd.DataFrame.from_records(records, columns=cols.split(',') + [count_column_name])
        return df


class Connection(sqlite3.Connection, SqlBase):
    """
    DDL - 数据定义语言
        CREATE	创建一个新的表，一个表的视图，或者数据库中的其他对象。
        ALTER	修改数据库中的某个已有的数据库对象，比如一个表。
        DROP	删除整个表，或者表的视图，或者数据库中的其他对象。
    DML - 数据操作语言
        INSERT	创建一条记录。
        UPDATE	修改记录。
        DELETE	删除记录。
    DQL - 数据查询语言
        SELECT	从一个或多个表中检索某些记录。
    """

    def has_table(self, table_name):
        res = self.execute(f"SELECT name FROM sqlite_master WHERE type='table' AND name='{table_name}'").fetchone()
        return bool(res)

    @classmethod
    def autotype(cls, val):
        if isinstance(val, str):
            return 'text'
        elif isinstance(val, (int, bool)):
            return 'integer'
        elif isinstance(val, float):
            return 'real'
        else:  # 其他dict、list等类型，可以用json.dumps或str转文本存储
            return 'text'

    @classmethod
    def cvt_type(cls, val):
        if isinstance(val, (dict, list, tuple)):
            val = json.dumps(val, ensure_ascii=False)
        return val

    @classmethod
    def cvt_types(cls, vals):
        return [cls.cvt_type(v) for v in vals]

    def create_table(self, table_name, column_descs):
        """ 【DDL增】
        :param table_name:
        :param column_descs:
            str, 正常的列格式描述，例如 'c1 text, c2 blob'
            dict, k是列名，v是一个具体的值，供分析参考格式类型
        :return:
        """
        # 1 列数据格式智能分析
        if not isinstance(column_descs, str):
            descs = []
            for k, v in column_descs.items():
                t = self.autotype(v)
                descs.append(f'{k} {t}')
            column_descs = ','.join(descs)

        # 2 新建表格
        self.execute(f'CREATE TABLE {table_name}({column_descs})')

    def ensure_column(self, table_name, col_name, col_type='', *, col_ref_val=None):
        """ 【DDL改】添加字段
        :param table_name:
        :param col_name:
        :param col_type:
        :param col_ref_val: 可以通过具体的数值，智能判断列数据格式
        :return:
        """
        if col_ref_val:
            col_type = self.autotype(col_ref_val)
        if not col_type:
            col_type = 'text'

        try:
            self.execute(f'ALTER TABLE {table_name} ADD COLUMN {col_name} {col_type}')
        except sqlite3.OperationalError as e:
            # 重复插入不报错
            if e.args[0] != f'duplicate column name: {col_name}':
                raise e

    def get_column_names(self, table_name):
        """ 【查】表格有哪些字段
        """
        names = []
        for item in self.execute(f'PRAGMA table_info({table_name})'):
            names.append(item[1])
        return names

    def insert(self, table_name, cols, if_exists='IGNORE'):
        """ 【增】插入新数据
        :param table_name:
        :param cols: 一般是用字典表示的要插入的值
        :param if_exists: 如果已存在的处理策略
            IGNORE，跳过
            REPLACE，替换
        :return:

        注意加了 OR IGNORE，支持重复数据自动忽略插入
        """
        ks = ','.join(cols.keys())
        vs = ','.join('?' * (len(cols.keys())))
        self.execute(f'INSERT OR {if_exists} INTO {table_name}({ks}) VALUES ({vs})', self.cvt_types(cols.values()))

    def update(self, table_name, cols, where):
        """ 【改】更新数据
        :param table_name:
        :param dict cols: 要更新的字段及值
        :param dict where: 怎么匹配到对应记录
        :return:
        """
        kvs = ','.join([f'{k}=?' for k in cols.keys()])
        ops = ' AND '.join([f'{k}=?' for k in where.keys()])
        vals = list(cols.values()) + list(where.values())
        self.execute(f'UPDATE {table_name} SET {kvs} WHERE {ops}', self.cvt_types(vals))

    def exec_nametuple(self, *args, **kwargs):
        cur = self.cursor()
        cur.row_factory = sqlite3.Row
        return cur.execute(*args, **kwargs)

    def exec_dict(self, *args, **kwargs):
        """ execute基础上，改成返回值为dict类型 """

        def dict_factory(cursor, row):
            d = {}
            for idx, col in enumerate(cursor.description):
                d[col[0]] = row[idx]
            return d

        cur = self.cursor()
        cur.row_factory = dict_factory
        return cur.execute(*args, **kwargs)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# @Author : 陈坤泽
# @Email  : 877362867@qq.com
# @Date   : 2021/03/24 17:03

""" 数据处理相关工具 """

from pyxlpr.data.imtextline import *
from pyxlpr.data.removeline import *
from pyxlpr.data.datacls import *



# copyright (c) 2020 PaddlePaddle Authors. All Rights Reserve.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
This code is refer from:
https://github.com/ayumiymk/aster.pytorch/blob/master/lib/models/tps_spatial_transformer.py
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import math
import paddle
from paddle import nn, ParamAttr
from paddle.nn import functional as F
import numpy as np
import itertools


def grid_sample(input, grid, canvas=None):
    input.stop_gradient = False
    output = F.grid_sample(input, grid)
    if canvas is None:
        return output
    else:
        input_mask = paddle.ones(shape=input.shape)
        output_mask = F.grid_sample(input_mask, grid)
        padded_output = output * output_mask + canvas * (1 - output_mask)
        return padded_output


# phi(x1, x2) = r^2 * log(r), where r = ||x1 - x2||_2
def compute_partial_repr(input_points, control_points):
    N = input_points.shape[0]
    M = control_points.shape[0]
    pairwise_diff = paddle.reshape(
        input_points, shape=[N, 1, 2]) - paddle.reshape(
            control_points, shape=[1, M, 2])
    # original implementation, very slow
    # pairwise_dist = torch.sum(pairwise_diff ** 2, dim = 2) # square of distance
    pairwise_diff_square = pairwise_diff * pairwise_diff
    pairwise_dist = pairwise_diff_square[:, :, 0] + pairwise_diff_square[:, :,
                                                                         1]
    repr_matrix = 0.5 * pairwise_dist * paddle.log(pairwise_dist)
    # fix numerical error for 0 * log(0), substitute all nan with 0
    mask = np.array(repr_matrix != repr_matrix)
    repr_matrix[mask] = 0
    return repr_matrix


# output_ctrl_pts are specified, according to our task.
def build_output_control_points(num_control_points, margins):
    margin_x, margin_y = margins
    num_ctrl_pts_per_side = num_control_points // 2
    ctrl_pts_x = np.linspace(margin_x, 1.0 - margin_x, num_ctrl_pts_per_side)
    ctrl_pts_y_top = np.ones(num_ctrl_pts_per_side) * margin_y
    ctrl_pts_y_bottom = np.ones(num_ctrl_pts_per_side) * (1.0 - margin_y)
    ctrl_pts_top = np.stack([ctrl_pts_x, ctrl_pts_y_top], axis=1)
    ctrl_pts_bottom = np.stack([ctrl_pts_x, ctrl_pts_y_bottom], axis=1)
    output_ctrl_pts_arr = np.concatenate(
        [ctrl_pts_top, ctrl_pts_bottom], axis=0)
    output_ctrl_pts = paddle.to_tensor(output_ctrl_pts_arr)
    return output_ctrl_pts


class TPSSpatialTransformer(nn.Layer):
    def __init__(self,
                 output_image_size=None,
                 num_control_points=None,
                 margins=None):
        super(TPSSpatialTransformer, self).__init__()
        self.output_image_size = output_image_size
        self.num_control_points = num_control_points
        self.margins = margins

        self.target_height, self.target_width = output_image_size
        target_control_points = build_output_control_points(num_control_points,
                                                            margins)
        N = num_control_points

        # create padded kernel matrix
        forward_kernel = paddle.zeros(shape=[N + 3, N + 3])
        target_control_partial_repr = compute_partial_repr(
            target_control_points, target_control_points)
        target_control_partial_repr = paddle.cast(target_control_partial_repr,
                                                  forward_kernel.dtype)
        forward_kernel[:N, :N] = target_control_partial_repr
        forward_kernel[:N, -3] = 1
        forward_kernel[-3, :N] = 1
        target_control_points = paddle.cast(target_control_points,
                                            forward_kernel.dtype)
        forward_kernel[:N, -2:] = target_control_points
        forward_kernel[-2:, :N] = paddle.transpose(
            target_control_points, perm=[1, 0])
        # compute inverse matrix
        inverse_kernel = paddle.inverse(forward_kernel)

        # create target cordinate matrix
        HW = self.target_height * self.target_width
        target_coordinate = list(
            itertools.product(
                range(self.target_height), range(self.target_width)))
        target_coordinate = paddle.to_tensor(target_coordinate)  # HW x 2
        Y, X = paddle.split(
            target_coordinate, target_coordinate.shape[1], axis=1)
        Y = Y / (self.target_height - 1)
        X = X / (self.target_width - 1)
        target_coordinate = paddle.concat(
            [X, Y], axis=1)  # convert from (y, x) to (x, y)
        target_coordinate_partial_repr = compute_partial_repr(
            target_coordinate, target_control_points)
        target_coordinate_repr = paddle.concat(
            [
                target_coordinate_partial_repr, paddle.ones(shape=[HW, 1]),
                target_coordinate
            ],
            axis=1)

        # register precomputed matrices
        self.inverse_kernel = inverse_kernel
        self.padding_matrix = paddle.zeros(shape=[3, 2])
        self.target_coordinate_repr = target_coordinate_repr
        self.target_control_points = target_control_points

    def forward(self, input, source_control_points):
        assert source_control_points.ndimension() == 3
        assert source_control_points.shape[1] == self.num_control_points
        assert source_control_points.shape[2] == 2
        batch_size = paddle.shape(source_control_points)[0]

        self.padding_matrix = paddle.expand(
            self.padding_matrix, shape=[batch_size, 3, 2])
        Y = paddle.concat([source_control_points, self.padding_matrix], 1)
        mapping_matrix = paddle.matmul(self.inverse_kernel, Y)
        source_coordinate = paddle.matmul(self.target_coordinate_repr,
                                          mapping_matrix)

        grid = paddle.reshape(
            source_coordinate,
            shape=[-1, self.target_height, self.target_width, 2])
        grid = paddle.clip(grid, 0,
                           1)  # the source_control_points may be out of [0, 1].
        # the input to grid_sample is normalized [-1, 1], but what we get is [0, 1]
        grid = 2.0 * grid - 1.0
        output_maps = grid_sample(input, grid, canvas=None)
        return output_maps, source_coordinate



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# @Author : 陈坤泽
# @Email  : 877362867@qq.com
# @Date   : 2021/06/06 10:51

from pyxllib.prog.newbie import round_int


def vector_compare(x, y):
    """
    :param x: 一个数值向量
    :param y: 一个数值向量，理论上长度要跟x相同
    :return: 返回<、=、>、?
        =，各个位置的数值都相同
        <，x各个位置的值都≤y，且至少有一处值<y
        >，x各个位置的值都≥，且至少有一处值>y
        ?，其他情况，x、y各有优劣

    >>> vector_compare([1, 2, 3], [1, 3, 3])
    '<'
    >>> vector_compare([1, 2, 3], [1, 1, 4])
    '?'
    >>> vector_compare({'a': 1, 'b': 2}, {'a': 1, 'b': 2, 'c': 0})
    '='
    >>> vector_compare({'b': 3, 'a': 1}, {'a': 1, 'b': 2, 'c': 0})
    '>'
    >>> vector_compare({'b': 3, 'a': 1}, {'a': 1, 'b': 2, 'c': 1})
    '?'
    """
    # 1 不同数据类型处理
    if isinstance(x, dict) and isinstance(y, dict):
        # 字典比较比较特别，要先做个映射
        keys = set(x.keys()) | set(y.keys())
        a = [x.get(k, 0) - y.get(k, 0) for k in keys]
    else:
        a = [i - j for i, j in zip(x, y)]

    # 2 算法
    n, m = max(a), min(a)
    if n == m == 0:
        return '='
    elif n <= 0:
        return '<'
    elif m >= 0:
        return '>'
    else:
        return '?'


def round_unit(x, unit):
    """ 按特定单位量对x取倍率

    round_int偏向于工程代码简化，round_unit偏向算法，功能不太一样，所以分组不同

    Args:
        x: 原值
        unit: 单位量

    Returns: 新值，是unit的整数倍

    >>> round_unit(1.2, 0.5)
    1.0
    >>> round_unit(1.6, 0.5)
    1.5
    >>> round_unit(7, 5)
    5
    >>> round_unit(13, 5)
    15
    """
    return round_int(x / unit) * unit


def int2excel_col_name(d):
    """
    >>> int2excel_col_name(1)
    'A'
    >>> int2excel_col_name(28)
    'AB'
    >>> int2excel_col_name(100)
    'CV'
    """
    s = []
    while d:
        t = (d - 1) % 26
        s.append(chr(65 + t))
        d = (d - 1) // 26
    return ''.join(reversed(s))


def excel_col_name2int(s):
    """
    >>> excel_col_name2int('A')
    1
    >>> excel_col_name2int('AA')
    27
    >>> excel_col_name2int('AB')
    28
    """
    d = 0
    for ch in s:
        d = d * 26 + (ord(ch.upper()) - 64)
    return d


def gentuple(n, tag):
    """ 有点类似range函数，但生成的数列更加灵活

    :param n:
        数组长度
    :param tag:
        int类型，从指定数字开始编号
            0，从0开始编号
            1，从1开始编号
        'A'，用Excel的形式编号
        tuple，按枚举值循环显示
            ('A', 'B')：循环使用A、B编号

    >>> gentuple(4, 'A')
    ('A', 'B', 'C', 'D')
    """
    a = [''] * n
    if isinstance(tag, int):
        for i in range(n):
            a[i] = i + tag
    elif tag == 'A':
        a = tuple(map(lambda x: int2excel_col_name(x + 1), range(n)))
    elif isinstance(tag, (list, tuple)):
        k = len(tag)
        a = tuple(map(lambda x: tag[x % k], range(n)))
    return a


def cvsecs(time):
    """ 从moviepy抄来的，任何时间格式转秒数的功能
    Will convert any time into seconds.

    If the type of `time` is not valid,
    it's returned as is.

    Here are the accepted formats::

    >>> cvsecs(15.4)   # seconds
    15.4
    >>> cvsecs((1, 21.5))   # (min,sec)
    81.5
    >>> cvsecs((1, 1, 2))   # (hr, min, sec)
    3662
    >>> cvsecs('01:01:33.045')
    3693.045
    >>> cvsecs('01:01:33,5')    # coma works too
    3693.5
    >>> cvsecs('1:33,5')    # only minutes and secs
    99.5
    >>> cvsecs('33.5')      # only secs
    33.5
    """
    factors = (1, 60, 3600)

    if isinstance(time, str):
        time = [float(f.replace(',', '.')) for f in time.split(':')]

    if not isinstance(time, (tuple, list)):
        return time

    return sum(mult * part for mult, part in zip(factors, reversed(time)))



# copyright (c) 2020 PaddlePaddle Authors. All Rights Reserve.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

from paddle import optimizer as optim


class Momentum(object):
    """
    Simple Momentum optimizer with velocity state.
    Args:
        learning_rate (float|Variable) - The learning rate used to update parameters.
            Can be a float value or a Variable with one float value as data element.
        momentum (float) - Momentum factor.
        regularization (WeightDecayRegularizer, optional) - The strategy of regularization.
    """

    def __init__(self,
                 learning_rate,
                 momentum,
                 weight_decay=None,
                 grad_clip=None,
                 **args):
        super(Momentum, self).__init__()
        self.learning_rate = learning_rate
        self.momentum = momentum
        self.weight_decay = weight_decay
        self.grad_clip = grad_clip

    def __call__(self, parameters):
        opt = optim.Momentum(
            learning_rate=self.learning_rate,
            momentum=self.momentum,
            weight_decay=self.weight_decay,
            grad_clip=self.grad_clip,
            parameters=parameters)
        return opt


class Adam(object):
    def __init__(self,
                 learning_rate=0.001,
                 beta1=0.9,
                 beta2=0.999,
                 epsilon=1e-08,
                 parameter_list=None,
                 weight_decay=None,
                 grad_clip=None,
                 name=None,
                 lazy_mode=False,
                 **kwargs):
        self.learning_rate = learning_rate
        self.beta1 = beta1
        self.beta2 = beta2
        self.epsilon = epsilon
        self.parameter_list = parameter_list
        self.learning_rate = learning_rate
        self.weight_decay = weight_decay
        self.grad_clip = grad_clip
        self.name = name
        self.lazy_mode = lazy_mode

    def __call__(self, parameters):
        opt = optim.Adam(
            learning_rate=self.learning_rate,
            beta1=self.beta1,
            beta2=self.beta2,
            epsilon=self.epsilon,
            weight_decay=self.weight_decay,
            grad_clip=self.grad_clip,
            name=self.name,
            lazy_mode=self.lazy_mode,
            parameters=parameters)
        return opt


class RMSProp(object):
    """
    Root Mean Squared Propagation (RMSProp) is an unpublished, adaptive learning rate method.
    Args:
        learning_rate (float|Variable) - The learning rate used to update parameters.
            Can be a float value or a Variable with one float value as data element.
        momentum (float) - Momentum factor.
        rho (float) - rho value in equation.
        epsilon (float) - avoid division by zero, default is 1e-6.
        regularization (WeightDecayRegularizer, optional) - The strategy of regularization.
    """

    def __init__(self,
                 learning_rate,
                 momentum=0.0,
                 rho=0.95,
                 epsilon=1e-6,
                 weight_decay=None,
                 grad_clip=None,
                 **args):
        super(RMSProp, self).__init__()
        self.learning_rate = learning_rate
        self.momentum = momentum
        self.rho = rho
        self.epsilon = epsilon
        self.weight_decay = weight_decay
        self.grad_clip = grad_clip

    def __call__(self, parameters):
        opt = optim.RMSProp(
            learning_rate=self.learning_rate,
            momentum=self.momentum,
            rho=self.rho,
            epsilon=self.epsilon,
            weight_decay=self.weight_decay,
            grad_clip=self.grad_clip,
            parameters=parameters)
        return opt


class Adadelta(object):
    def __init__(self,
                 learning_rate=0.001,
                 epsilon=1e-08,
                 rho=0.95,
                 parameter_list=None,
                 weight_decay=None,
                 grad_clip=None,
                 name=None,
                 **kwargs):
        self.learning_rate = learning_rate
        self.epsilon = epsilon
        self.rho = rho
        self.parameter_list = parameter_list
        self.learning_rate = learning_rate
        self.weight_decay = weight_decay
        self.grad_clip = grad_clip
        self.name = name

    def __call__(self, parameters):
        opt = optim.Adadelta(
            learning_rate=self.learning_rate,
            epsilon=self.epsilon,
            rho=self.rho,
            weight_decay=self.weight_decay,
            grad_clip=self.grad_clip,
            name=self.name,
            parameters=parameters)
        return opt



#copyright (c) 2021 PaddlePaddle Authors. All Rights Reserve.
#
#Licensed under the Apache License, Version 2.0 (the "License");
#you may not use this file except in compliance with the License.
#You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
#Unless required by applicable law or agreed to in writing, software
#distributed under the License is distributed on an "AS IS" BASIS,
#WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#See the License for the specific language governing permissions and
#limitations under the License.

import paddle
import paddle.nn as nn
import numpy as np
import cv2

from .rec_ctc_loss import CTCLoss
from .basic_loss import DMLLoss
from .basic_loss import DistanceLoss
from .det_db_loss import DBLoss
from .det_basic_loss import BalanceLoss, MaskL1Loss, DiceLoss


def _sum_loss(loss_dict):
    if "loss" in loss_dict.keys():
        return loss_dict
    else:
        loss_dict["loss"] = 0.
        for k, value in loss_dict.items():
            if k == "loss":
                continue
            else:
                loss_dict["loss"] += value
        return loss_dict


class DistillationDMLLoss(DMLLoss):
    """
    """

    def __init__(self,
                 model_name_pairs=[],
                 act=None,
                 use_log=False,
                 key=None,
                 maps_name=None,
                 name="dml"):
        super().__init__(act=act, use_log=use_log)
        assert isinstance(model_name_pairs, list)
        self.key = key
        self.model_name_pairs = self._check_model_name_pairs(model_name_pairs)
        self.name = name
        self.maps_name = self._check_maps_name(maps_name)

    def _check_model_name_pairs(self, model_name_pairs):
        if not isinstance(model_name_pairs, list):
            return []
        elif isinstance(model_name_pairs[0], list) and isinstance(
                model_name_pairs[0][0], str):
            return model_name_pairs
        else:
            return [model_name_pairs]

    def _check_maps_name(self, maps_name):
        if maps_name is None:
            return None
        elif type(maps_name) == str:
            return [maps_name]
        elif type(maps_name) == list:
            return [maps_name]
        else:
            return None

    def _slice_out(self, outs):
        new_outs = {}
        for k in self.maps_name:
            if k == "thrink_maps":
                new_outs[k] = outs[:, 0, :, :]
            elif k == "threshold_maps":
                new_outs[k] = outs[:, 1, :, :]
            elif k == "binary_maps":
                new_outs[k] = outs[:, 2, :, :]
            else:
                continue
        return new_outs

    def forward(self, predicts, batch):
        loss_dict = dict()
        for idx, pair in enumerate(self.model_name_pairs):
            out1 = predicts[pair[0]]
            out2 = predicts[pair[1]]
            if self.key is not None:
                out1 = out1[self.key]
                out2 = out2[self.key]

            if self.maps_name is None:
                loss = super().forward(out1, out2)
                if isinstance(loss, dict):
                    for key in loss:
                        loss_dict["{}_{}_{}_{}".format(key, pair[0], pair[1],
                                                       idx)] = loss[key]
                else:
                    loss_dict["{}_{}".format(self.name, idx)] = loss
            else:
                outs1 = self._slice_out(out1)
                outs2 = self._slice_out(out2)
                for _c, k in enumerate(outs1.keys()):
                    loss = super().forward(outs1[k], outs2[k])
                    if isinstance(loss, dict):
                        for key in loss:
                            loss_dict["{}_{}_{}_{}_{}".format(key, pair[
                                0], pair[1], self.maps_name, idx)] = loss[key]
                    else:
                        loss_dict["{}_{}_{}".format(self.name, self.maps_name[
                            _c], idx)] = loss

        loss_dict = _sum_loss(loss_dict)

        return loss_dict


class DistillationCTCLoss(CTCLoss):
    def __init__(self, model_name_list=[], key=None, name="loss_ctc"):
        super().__init__()
        self.model_name_list = model_name_list
        self.key = key
        self.name = name

    def forward(self, predicts, batch):
        loss_dict = dict()
        for idx, model_name in enumerate(self.model_name_list):
            out = predicts[model_name]
            if self.key is not None:
                out = out[self.key]
            loss = super().forward(out, batch)
            if isinstance(loss, dict):
                for key in loss:
                    loss_dict["{}_{}_{}".format(self.name, model_name,
                                                idx)] = loss[key]
            else:
                loss_dict["{}_{}".format(self.name, model_name)] = loss
        return loss_dict


class DistillationDBLoss(DBLoss):
    def __init__(self,
                 model_name_list=[],
                 balance_loss=True,
                 main_loss_type='DiceLoss',
                 alpha=5,
                 beta=10,
                 ohem_ratio=3,
                 eps=1e-6,
                 name="db",
                 **kwargs):
        super().__init__()
        self.model_name_list = model_name_list
        self.name = name
        self.key = None

    def forward(self, predicts, batch):
        loss_dict = {}
        for idx, model_name in enumerate(self.model_name_list):
            out = predicts[model_name]
            if self.key is not None:
                out = out[self.key]
            loss = super().forward(out, batch)

            if isinstance(loss, dict):
                for key in loss.keys():
                    if key == "loss":
                        continue
                    name = "{}_{}_{}".format(self.name, model_name, key)
                    loss_dict[name] = loss[key]
            else:
                loss_dict["{}_{}".format(self.name, model_name)] = loss

        loss_dict = _sum_loss(loss_dict)
        return loss_dict


class DistillationDilaDBLoss(DBLoss):
    def __init__(self,
                 model_name_pairs=[],
                 key=None,
                 balance_loss=True,
                 main_loss_type='DiceLoss',
                 alpha=5,
                 beta=10,
                 ohem_ratio=3,
                 eps=1e-6,
                 name="dila_dbloss"):
        super().__init__()
        self.model_name_pairs = model_name_pairs
        self.name = name
        self.key = key

    def forward(self, predicts, batch):
        loss_dict = dict()
        for idx, pair in enumerate(self.model_name_pairs):
            stu_outs = predicts[pair[0]]
            tch_outs = predicts[pair[1]]
            if self.key is not None:
                stu_preds = stu_outs[self.key]
                tch_preds = tch_outs[self.key]

            stu_shrink_maps = stu_preds[:, 0, :, :]
            stu_binary_maps = stu_preds[:, 2, :, :]

            # dilation to teacher prediction
            dilation_w = np.array([[1, 1], [1, 1]])
            th_shrink_maps = tch_preds[:, 0, :, :]
            th_shrink_maps = th_shrink_maps.numpy() > 0.3  # thresh = 0.3 
            dilate_maps = np.zeros_like(th_shrink_maps).astype(np.float32)
            for i in range(th_shrink_maps.shape[0]):
                dilate_maps[i] = cv2.dilate(
                    th_shrink_maps[i, :, :].astype(np.uint8), dilation_w)
            th_shrink_maps = paddle.to_tensor(dilate_maps)

            label_threshold_map, label_threshold_mask, label_shrink_map, label_shrink_mask = batch[
                1:]

            # calculate the shrink map loss
            bce_loss = self.alpha * self.bce_loss(
                stu_shrink_maps, th_shrink_maps, label_shrink_mask)
            loss_binary_maps = self.dice_loss(stu_binary_maps, th_shrink_maps,
                                              label_shrink_mask)

            # k = f"{self.name}_{pair[0]}_{pair[1]}"
            k = "{}_{}_{}".format(self.name, pair[0], pair[1])
            loss_dict[k] = bce_loss + loss_binary_maps

        loss_dict = _sum_loss(loss_dict)
        return loss_dict


class DistillationDistanceLoss(DistanceLoss):
    """
    """

    def __init__(self,
                 mode="l2",
                 model_name_pairs=[],
                 key=None,
                 name="loss_distance",
                 **kargs):
        super().__init__(mode=mode, **kargs)
        assert isinstance(model_name_pairs, list)
        self.key = key
        self.model_name_pairs = model_name_pairs
        self.name = name + "_l2"

    def forward(self, predicts, batch):
        loss_dict = dict()
        for idx, pair in enumerate(self.model_name_pairs):
            out1 = predicts[pair[0]]
            out2 = predicts[pair[1]]
            if self.key is not None:
                out1 = out1[self.key]
                out2 = out2[self.key]
            loss = super().forward(out1, out2)
            if isinstance(loss, dict):
                for key in loss:
                    loss_dict["{}_{}_{}".format(self.name, key, idx)] = loss[
                        key]
            else:
                loss_dict["{}_{}_{}_{}".format(self.name, pair[0], pair[1],
                                               idx)] = loss
        return loss_dict



# copyright (c) 2020 PaddlePaddle Authors. All Rights Reserve.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from paddle import nn
import paddle


class MTB(nn.Layer):
    def __init__(self, cnn_num, in_channels):
        super(MTB, self).__init__()
        self.block = nn.Sequential()
        self.out_channels = in_channels
        self.cnn_num = cnn_num
        if self.cnn_num == 2:
            for i in range(self.cnn_num):
                self.block.add_sublayer(
                    'conv_{}'.format(i),
                    nn.Conv2D(
                        in_channels=in_channels
                        if i == 0 else 32 * (2**(i - 1)),
                        out_channels=32 * (2**i),
                        kernel_size=3,
                        stride=2,
                        padding=1))
                self.block.add_sublayer('relu_{}'.format(i), nn.ReLU())
                self.block.add_sublayer('bn_{}'.format(i),
                                        nn.BatchNorm2D(32 * (2**i)))

    def forward(self, images):
        x = self.block(images)
        if self.cnn_num == 2:
            # (b, w, h, c)
            x = paddle.transpose(x, [0, 3, 2, 1])
            x_shape = paddle.shape(x)
            x = paddle.reshape(
                x, [x_shape[0], x_shape[1], x_shape[2] * x_shape[3]])
        return x
